# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import faulthandler
import logging
import logging.handlers
import os
import signal
import sys
import threading
import time
import traceback
from collections import defaultdict
from contextlib import ExitStack
from datetime import datetime
from queue import Empty, Queue
from typing import Any, Dict, Iterable, List, Optional, Sequence, Set, Tuple

import grpc
from grpc_reflection.v1alpha import reflection

from buildgrid._protos.buildgrid.v2.monitoring_pb2 import LogRecord
from buildgrid.server.actioncache.service import ActionCacheService
from buildgrid.server.bots.service import BotsService
from buildgrid.server.build_events.service import PublishBuildEventService, QueryBuildEventsService
from buildgrid.server.capabilities.instance import CapabilitiesInstance
from buildgrid.server.capabilities.service import CapabilitiesService
from buildgrid.server.cas.service import ByteStreamService, ContentAddressableStorageService
from buildgrid.server.context import instance_context
from buildgrid.server.controller import ExecutionController
from buildgrid.server.enums import LeaseState, LogRecordLevel, MetricCategories, OperationStage
from buildgrid.server.exceptions import PermissionDeniedError
from buildgrid.server.execution.service import ExecutionService
from buildgrid.server.introspection.service import IntrospectionService
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.metrics_names import METRIC
from buildgrid.server.metrics_utils import publish_gauge_metric
from buildgrid.server.monitoring import get_monitoring_bus
from buildgrid.server.operations.service import OperationsService
from buildgrid.server.scheduler import Scheduler, SchedulerMetrics
from buildgrid.server.servicer import Instance, InstancedServicer
from buildgrid.server.settings import LOG_RECORD_FORMAT, MIN_THREAD_POOL_SIZE, MONITORING_PERIOD, SHUTDOWN_ALARM_DELAY
from buildgrid.server.threading import ContextThreadPoolExecutor, ContextWorker
from buildgrid.server.types import OnServerStartCallback, PortAssignedCallback

LOGGER = buildgrid_logger(__name__)


def load_tls_server_credentials(
    server_key: Optional[str] = None, server_cert: Optional[str] = None, client_certs: Optional[str] = None
) -> Optional[grpc.ServerCredentials]:
    """Looks-up and loads TLS server gRPC credentials.

    Every private and public keys are expected to be PEM-encoded.

    Args:
        server_key(str): private server key file path.
        server_cert(str): public server certificate file path.
        client_certs(str): public client certificates file path.

    Returns:
        :obj:`ServerCredentials`: The credentials for use for a
        TLS-encrypted gRPC server channel.
    """
    if not server_key or not os.path.exists(server_key):
        return None

    if not server_cert or not os.path.exists(server_cert):
        return None

    with open(server_key, "rb") as f:
        server_key_pem = f.read()
    with open(server_cert, "rb") as f:
        server_cert_pem = f.read()

    if client_certs and os.path.exists(client_certs):
        with open(client_certs, "rb") as f:
            client_certs_pem = f.read()
    else:
        client_certs_pem = None
        client_certs = None

    credentials = grpc.ssl_server_credentials(
        [(server_key_pem, server_cert_pem)], root_certificates=client_certs_pem, require_client_auth=bool(client_certs)
    )

    # TODO: Fix this (missing stubs?) "ServerCredentials" has no attribute
    credentials.server_key = server_key  # type: ignore[attr-defined]
    credentials.server_cert = server_cert  # type: ignore[attr-defined]
    credentials.client_certs = client_certs  # type: ignore[attr-defined]

    return credentials


class Server:
    """Creates a BuildGrid server instance.

    The :class:`Server` class binds together all the gRPC services.
    """

    def __init__(
        self,
        server_reflection: bool,
        grpc_compression: grpc.Compression,
        is_instrumented: bool,
        grpc_server_options: Optional[Sequence[Tuple[str, Any]]],
        max_workers: Optional[int],
        monitoring_period: float = MONITORING_PERIOD,
    ):
        self._stack = ExitStack()

        self._server_reflection = server_reflection
        self._grpc_compression = grpc_compression
        self._is_instrumented = is_instrumented
        self._grpc_server_options = grpc_server_options

        self._action_cache_service = ActionCacheService()
        self._bots_service = BotsService()
        self._bytestream_service = ByteStreamService()
        self._capabilities_service = CapabilitiesService()
        self._cas_service = ContentAddressableStorageService()
        self._execution_service = ExecutionService()
        self._operations_service = OperationsService()
        self._introspection_service = IntrospectionService()

        # Special cases
        self._build_events_service = PublishBuildEventService()
        self._query_build_events_service = QueryBuildEventsService()

        self._schedulers: Dict[str, Set[Scheduler]] = defaultdict(set)

        self._ports: List[Tuple[str, Optional[Dict[str, str]]]] = []
        self._port_map: Dict[str, int] = {}

        self._logging_queue: Queue[Any] = Queue()
        self._monitoring_period = monitoring_period

        if max_workers is None:
            # Use max_workers default from Python 3.4+
            max_workers = max(MIN_THREAD_POOL_SIZE, (os.cpu_count() or 1) * 5)

        elif max_workers < MIN_THREAD_POOL_SIZE:
            LOGGER.warning(
                "Specified thread-limit is too small, bumping it.",
                tags=dict(requested_thread_limit=max_workers, new_thread_limit=MIN_THREAD_POOL_SIZE),
            )
            # Enforce a minumun for max_workers
            max_workers = MIN_THREAD_POOL_SIZE

        self._max_grpc_workers = max_workers

    def register_instance(self, instance_name: str, instance: Instance) -> None:
        """
        Register an instance with the server. Handled the logic of mapping instances to the
        correct servicer container.

        Args:
            instance_name (str): The name of the instance.

            instance (Instance): The instance implementation.
        """

        # Special case to handle the ExecutionController which combines the service interfaces.
        if isinstance(instance, ExecutionController):
            if bots_interface := instance.bots_interface:
                self.register_instance(instance_name, bots_interface)
            if execution_instance := instance.execution_instance:
                self.register_instance(instance_name, execution_instance)
            if operations_instance := instance.operations_instance:
                self.register_instance(instance_name, operations_instance)

        elif action_instance := self._action_cache_service.cast(instance):
            self._action_cache_service.add_instance(instance_name, action_instance)
            capabilities = self._capabilities_service.instances.setdefault(instance_name, CapabilitiesInstance())
            capabilities.add_action_cache_instance(action_instance)

        elif bots_instance := self._bots_service.cast(instance):
            self._bots_service.add_instance(instance_name, bots_instance)
            self._schedulers[instance_name].add(bots_instance.scheduler)

        elif bytestream_instance := self._bytestream_service.cast(instance):
            self._bytestream_service.add_instance(instance_name, bytestream_instance)

        elif cas_instance := self._cas_service.cast(instance):
            self._cas_service.add_instance(instance_name, cas_instance)
            capabilities = self._capabilities_service.instances.setdefault(instance_name, CapabilitiesInstance())
            capabilities.add_cas_instance(cas_instance)

        elif execution_instance := self._execution_service.cast(instance):
            self._execution_service.add_instance(instance_name, execution_instance)
            self._schedulers[instance_name].add(execution_instance.scheduler)
            capabilities = self._capabilities_service.instances.setdefault(instance_name, CapabilitiesInstance())
            capabilities.add_execution_instance(execution_instance)

        elif operations_instance := self._operations_service.cast(instance):
            self._operations_service.add_instance(instance_name, operations_instance)

        elif introspection_instance := self._introspection_service.cast(instance):
            self._introspection_service.add_instance(instance_name, introspection_instance)

        # The Build Events Services have no support for instance names, so this
        # is a bit of a special case where the storage backend itself is the
        # trigger for creating the gRPC services.
        elif instance.SERVICE_NAME == "BuildEvents":
            self._build_events_service.add_instance("", instance)  # type: ignore[arg-type]
            self._query_build_events_service.add_instance("", instance)  # type: ignore[arg-type]

        else:
            raise ValueError(f"Instance of type {type(instance)} not supported by {type(self)}")

    @property
    def _services(self) -> Iterable[InstancedServicer[Any]]:
        return (
            self._action_cache_service,
            self._bots_service,
            self._bytestream_service,
            self._capabilities_service,
            self._cas_service,
            self._execution_service,
            self._operations_service,
            self._introspection_service,
            # Special cases
            self._build_events_service,
            self._query_build_events_service,
        )

    def add_port(self, address: str, credentials: Optional[Dict[str, str]]) -> None:
        """Adds a port to the server.

        Must be called before the server starts. If a credentials object exists,
        it will make a secure port.

        Args:
            address (str): The address with port number.
            credentials (:obj:`grpc.ChannelCredentials`): Credentials object.
        """
        self._ports.append((address, credentials))

    def start(
        self,
        *,
        on_server_start_cb: Optional[OnServerStartCallback] = None,
        port_assigned_callback: Optional[PortAssignedCallback] = None,
        run_forever: bool = True,
    ) -> None:
        """Starts the BuildGrid server.

        BuildGrid server startup consists of 3 stages,

        1. Starting logging and monitoring

        This step starts up the logging coroutine, the periodic status metrics
        coroutine, and the monitoring bus' publishing subprocess. Since this
        step involves forking, anything not fork-safe needs to be done *after*
        this step.

        2. Instantiate gRPC

        This step instantiates the gRPC server, and tells all the instances
        which have been attached to the server to instantiate their gRPC
        objects. It is also responsible for creating the various service
        objects and connecting them to the server and the instances.

        After this step, gRPC core is running and its no longer safe to fork
        the process.

        3. Start instances

        Several of BuildGrid's services use background threads that need to
        be explicitly started when BuildGrid starts up. Rather than doing
        this at configuration parsing time, this step provides a hook for
        services to start up in a more organised fashion.

        4. Start the gRPC server

        The final step is starting up the gRPC server. The callback passed in
        via ``on_server_start_cb`` is executed in this step once the server
        has started. After this point BuildGrid is ready to serve requests.

        The final thing done by this method is adding a ``SIGTERM`` handler
        which calls the ``Server.stop`` method to the event loop, and then
        that loop is started up using ``run_forever()``.

        Args:
            on_server_start_cb (Callable): Callback function to execute once
                the gRPC server has started up.
            port_assigned_callback (Callable): Callback function to execute
                once the gRPC server has started up. The mapping of addresses
                to ports is passed to this callback.

        """
        # 1. Start logging and monitoring
        self._stack.enter_context(
            ContextWorker(
                self._logging_worker,
                "ServerLogger",
                # Add a dummy value to the queue to unblock the get call.
                on_shutdown_requested=lambda: self._logging_queue.put(None),
            )
        )
        if self._is_instrumented:
            self._stack.enter_context(get_monitoring_bus())
            self._stack.enter_context(ContextWorker(self._state_monitoring_worker, "ServerMonitor"))

        # 2. Instantiate gRPC objects
        grpc_server = self.setup_grpc()

        # 3. Start background threads
        for service in self._services:
            self._stack.enter_context(service)

        # 4. Start the gRPC server.
        grpc_server.start()
        self._stack.callback(grpc_server.stop, None)

        if on_server_start_cb:
            on_server_start_cb()
        if port_assigned_callback:
            port_assigned_callback(port_map=self._port_map)

        # Add the stop handler and run the event loop
        if run_forever:
            grpc_server.wait_for_termination()

    def setup_grpc(self) -> grpc.Server:
        """Instantiate the gRPC objects.

        This creates the gRPC server, and causes the instances attached to
        this server to instantiate any gRPC channels they need. This also
        sets up the services which route to those instances, and sets up
        gRPC reflection.

        """
        LOGGER.info(
            "Setting up gRPC server.",
            tags=dict(
                maximum_concurrent_rpcs=self._max_grpc_workers,
                compression=self._grpc_compression,
                options=self._grpc_server_options,
            ),
        )

        grpc_server = grpc.server(
            ContextThreadPoolExecutor(self._max_grpc_workers, "gRPC_Executor", immediate_copy=True),
            maximum_concurrent_rpcs=self._max_grpc_workers,
            compression=self._grpc_compression,
            options=self._grpc_server_options,
        )

        # Add the requested ports to the gRPC server
        for address, credentials in self._ports:
            port_number = 0
            if credentials is not None:
                LOGGER.info("Adding secure connection.", tags=dict(address=address))
                server_key = credentials.get("tls-server-key")
                server_cert = credentials.get("tls-server-cert")
                client_certs = credentials.get("tls-client-certs")
                server_credentials = load_tls_server_credentials(
                    server_cert=server_cert, server_key=server_key, client_certs=client_certs
                )
                # TODO should this error out??
                if server_credentials:
                    port_number = grpc_server.add_secure_port(address, server_credentials)

            else:
                LOGGER.info("Adding insecure connection.", tags=dict(address=address))
                port_number = grpc_server.add_insecure_port(address)

            if not port_number:
                raise PermissionDeniedError("Unable to configure socket")

            self._port_map[address] = port_number

        for service in self._services:
            service.setup_grpc(grpc_server)

        if self._server_reflection:
            reflection_services = [service.FULL_NAME for service in self._services if service.enabled]
            LOGGER.info("Server reflection is enabled.", tags=dict(reflection_services=reflection_services))
            reflection.enable_server_reflection([reflection.SERVICE_NAME] + reflection_services, grpc_server)
        else:
            LOGGER.info("Server reflection is not enabled.")

        return grpc_server

    def stop(self, *args: Any, **kwargs: Any) -> None:
        LOGGER.info("Stopping BuildGrid server.")

        def alarm_handler(*args: Any, **kwargs: Any) -> None:
            LOGGER.warning(
                "Shutdown still ongoing after shutdown delay.",
                tags=dict(
                    shutdown_alarm_delay_seconds=SHUTDOWN_ALARM_DELAY, active_thread_count=threading.active_count()
                ),
            )
            for thread in threading.enumerate():
                LOGGER.warning(f"  - {thread.name}")
            LOGGER.warning("Traceback for all threads:")
            faulthandler.dump_traceback()

        LOGGER.debug("Setting alarm for delayed shutdown.")
        signal.signal(signal.SIGALRM, alarm_handler)
        signal.alarm(SHUTDOWN_ALARM_DELAY)

        self._stack.close()

    def _logging_worker(self, shutdown_requested: threading.Event) -> None:
        """Publishes log records to the monitoring bus."""

        logging_formatter = logging.Formatter(fmt=LOG_RECORD_FORMAT)
        logging_handler = logging.handlers.QueueHandler(self._logging_queue)

        # Setup the main logging handler:
        root_logger = logging.getLogger()

        for log_filter in root_logger.filters[:]:
            logging_handler.addFilter(log_filter)
            root_logger.removeFilter(log_filter)

        for log_handler in root_logger.handlers[:]:
            for log_filter in log_handler.filters[:]:
                logging_handler.addFilter(log_filter)
            root_logger.removeHandler(log_handler)
        root_logger.addHandler(logging_handler)

        def logging_worker() -> None:
            monitoring_bus = get_monitoring_bus()

            try:
                log_record = self._logging_queue.get(timeout=self._monitoring_period)
            except Empty:
                return
            if log_record is None:
                return

            # Print log records to stdout, if required:
            if not self._is_instrumented or not monitoring_bus.prints_records:
                record = logging_formatter.format(log_record)
                # TODO: Investigate if async write would be worth here.
                sys.stdout.write(f"{record}\n")
                sys.stdout.flush()

            # Emit a log record if server is instrumented:
            if self._is_instrumented:
                log_record_level = LogRecordLevel(int(log_record.levelno / 10))
                log_record_creation_time = datetime.fromtimestamp(log_record.created)
                # logging.LogRecord.extra must be a str to str dict:
                if "extra" in log_record.__dict__ and log_record.extra:
                    log_record_metadata = log_record.extra
                else:
                    log_record_metadata = None
                forged_record = self._forge_log_record(
                    domain=log_record.name,
                    level=log_record_level,
                    message=log_record.message,
                    creation_time=log_record_creation_time,
                    metadata=log_record_metadata,
                )
                monitoring_bus.send_record_nowait(forged_record)

        while not shutdown_requested.is_set():
            try:
                logging_worker()
            except Exception:
                # The thread shouldn't exit on exceptions, but output the exception so that
                # it can be found in the logs.
                #
                # Note, we DO NOT use `LOGGER` here, because we don't want to write
                # anything new to the logging queue in case the Exception isn't some transient
                # issue.
                try:
                    sys.stdout.write("Exception in logging worker\n")
                    sys.stdout.flush()
                    traceback.print_exc()
                except Exception:
                    # There's not a lot we can do at this point really.
                    pass

        if shutdown_requested.is_set():
            # Reset logging, so any logging after shutting down the logging worker
            # still gets written to stdout and the queue doesn't get any more logs
            stream_handler = logging.StreamHandler(stream=sys.stdout)
            stream_handler.setFormatter(logging_formatter)
            root_logger = logging.getLogger()

            for log_filter in root_logger.filters[:]:
                stream_handler.addFilter(log_filter)
                root_logger.removeFilter(log_filter)

            for log_handler in root_logger.handlers[:]:
                for log_filter in log_handler.filters[:]:
                    stream_handler.addFilter(log_filter)
                root_logger.removeHandler(log_handler)
            root_logger.addHandler(stream_handler)

            # Drain the log message queue
            while self._logging_queue.qsize() > 0:
                logging_worker()

    def _forge_log_record(
        self,
        *,
        domain: str,
        level: LogRecordLevel,
        message: str,
        creation_time: datetime,
        metadata: Optional[Dict[str, str]] = None,
    ) -> LogRecord:
        log_record = LogRecord()

        log_record.creation_timestamp.FromDatetime(creation_time)
        log_record.domain = domain
        log_record.level = level.value
        log_record.message = message
        if metadata is not None:
            log_record.metadata.update(metadata)

        return log_record

    def _state_monitoring_worker(self, shutdown_requested: threading.Event) -> None:
        """Periodically publishes state metrics to the monitoring bus."""
        while not shutdown_requested.is_set():
            start = time.time()
            try:
                if self._execution_service.enabled:
                    for instance_name in self._execution_service.instances:
                        self._publish_client_metrics_for_instance(instance_name)

                if self._bots_service.enabled:
                    for instance_name in self._bots_service.instances:
                        self._publish_bot_metrics_for_instance(instance_name)

                if self._schedulers:
                    for instance_name in self._schedulers:
                        self._publish_scheduler_metrics_for_instance(instance_name)
            except Exception:
                # The thread shouldn't exit on exceptions, but log at a severe enough level
                # that it doesn't get lost in logs
                LOGGER.exception("Exception while gathering state metrics.")

            end = time.time()
            shutdown_requested.wait(timeout=max(0, self._monitoring_period - (end - start)))

    def _publish_client_metrics_for_instance(self, instance_name: str) -> None:
        """Queries the number of clients connected for a given instance"""
        with instance_context(instance_name):
            n_clients = self._execution_service.query_connected_clients_for_instance(instance_name)
            publish_gauge_metric(METRIC.CONNECTIONS.CLIENT_COUNT, n_clients)

    def _publish_bot_metrics_for_instance(self, instance_name: str) -> None:
        with instance_context(instance_name):
            for bot_status, count in self._bots_service.count_bots_by_status(instance_name).items():
                publish_gauge_metric(METRIC.SCHEDULER.BOTS_COUNT, count, state=bot_status.name)

            n_workers = self._bots_service.query_connected_bots_for_instance(instance_name)
            publish_gauge_metric(METRIC.CONNECTIONS.WORKER_COUNT, n_workers)

    def _publish_scheduler_metrics_for_instance(self, instance_name: str) -> None:
        with instance_context(instance_name):
            # Since multiple schedulers may be active for this instance, but should
            # be using the same database, just use the first one
            scheduler_metrics: Optional[SchedulerMetrics] = None
            for scheduler in self._schedulers[instance_name]:
                scheduler_metrics = scheduler.get_metrics()
            if scheduler_metrics is None:
                return

            # Jobs
            for stage, n_jobs in scheduler_metrics[MetricCategories.JOBS.value].items():
                stage = OperationStage(stage)
                publish_gauge_metric(METRIC.SCHEDULER.JOB_COUNT, n_jobs, state=stage.name)
            # Leases
            for state, n_leases in scheduler_metrics[MetricCategories.LEASES.value].items():
                state = LeaseState(state)
                publish_gauge_metric(METRIC.SCHEDULER.LEASE_COUNT, n_leases, state=state.name)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2020 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import builtins
import collections.abc
import buildgrid._protos.google.devtools.build.v1.build_events_pb2
import google.protobuf.descriptor
import google.protobuf.duration_pb2
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class PublishLifecycleEventRequest(google.protobuf.message.Message):
    """Publishes 'lifecycle events' that update the high-level state of a build:
    - BuildEnqueued: When a build is scheduled.
    - InvocationAttemptStarted: When work for a build starts; there can be
        multiple invocations for a build (e.g. retries).
    - InvocationAttemptCompleted: When work for a build finishes.
    - BuildFinished: When a build is finished.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _ServiceLevel:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _ServiceLevelEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[PublishLifecycleEventRequest._ServiceLevel.ValueType], builtins.type):  # noqa: F821
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        NONINTERACTIVE: PublishLifecycleEventRequest._ServiceLevel.ValueType  # 0
        """Non-interactive builds can tolerate longer event latencies. This is the
        default ServiceLevel if callers do not specify one.
        """
        INTERACTIVE: PublishLifecycleEventRequest._ServiceLevel.ValueType  # 1
        """The events of an interactive build should be delivered with low latency."""

    class ServiceLevel(_ServiceLevel, metaclass=_ServiceLevelEnumTypeWrapper):
        """The service level of the build request. Backends only uses this value when
        the BuildEnqueued event is published to determine what level of service
        this build should receive.
        """

    NONINTERACTIVE: PublishLifecycleEventRequest.ServiceLevel.ValueType  # 0
    """Non-interactive builds can tolerate longer event latencies. This is the
    default ServiceLevel if callers do not specify one.
    """
    INTERACTIVE: PublishLifecycleEventRequest.ServiceLevel.ValueType  # 1
    """The events of an interactive build should be delivered with low latency."""

    SERVICE_LEVEL_FIELD_NUMBER: builtins.int
    BUILD_EVENT_FIELD_NUMBER: builtins.int
    STREAM_TIMEOUT_FIELD_NUMBER: builtins.int
    NOTIFICATION_KEYWORDS_FIELD_NUMBER: builtins.int
    PROJECT_ID_FIELD_NUMBER: builtins.int
    service_level: global___PublishLifecycleEventRequest.ServiceLevel.ValueType
    """The interactivity of this build."""
    @property
    def build_event(self) -> global___OrderedBuildEvent:
        """Required. The lifecycle build event. If this is a build tool event, the RPC
        will fail with INVALID_REQUEST.
        """
    @property
    def stream_timeout(self) -> google.protobuf.duration_pb2.Duration:
        """If the next event for this build or invocation (depending on the event
        type) hasn't been published after this duration from when {build_event}
        is written to BES, consider this stream expired. If this field is not set,
        BES backend will use its own default value.
        """
    @property
    def notification_keywords(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """Additional information about a build request. These are define by the event
        publishers, and the Build Event Service does not validate or interpret
        them. They are used while notifying internal systems of new builds and
        invocations if the OrderedBuildEvent.event type is
        BuildEnqueued/InvocationAttemptStarted.
        """
    project_id: builtins.str
    """Required. The project this build is associated with.
    This should match the project used for the initial call to
    PublishLifecycleEvent (containing a BuildEnqueued message).
    """
    def __init__(
        self,
        *,
        service_level: global___PublishLifecycleEventRequest.ServiceLevel.ValueType = ...,
        build_event: global___OrderedBuildEvent | None = ...,
        stream_timeout: google.protobuf.duration_pb2.Duration | None = ...,
        notification_keywords: collections.abc.Iterable[builtins.str] | None = ...,
        project_id: builtins.str = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["build_event", b"build_event", "stream_timeout", b"stream_timeout"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["build_event", b"build_event", "notification_keywords", b"notification_keywords", "project_id", b"project_id", "service_level", b"service_level", "stream_timeout", b"stream_timeout"]) -> None: ...

global___PublishLifecycleEventRequest = PublishLifecycleEventRequest

@typing_extensions.final
class PublishBuildToolEventStreamResponse(google.protobuf.message.Message):
    """States which event has been committed. Any failure to commit will cause
    RPC errors, hence not recorded by this proto.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STREAM_ID_FIELD_NUMBER: builtins.int
    SEQUENCE_NUMBER_FIELD_NUMBER: builtins.int
    @property
    def stream_id(self) -> buildgrid._protos.google.devtools.build.v1.build_events_pb2.StreamId:
        """The stream that contains this event."""
    sequence_number: builtins.int
    """The sequence number of this event that has been committed."""
    def __init__(
        self,
        *,
        stream_id: buildgrid._protos.google.devtools.build.v1.build_events_pb2.StreamId | None = ...,
        sequence_number: builtins.int = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["stream_id", b"stream_id"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["sequence_number", b"sequence_number", "stream_id", b"stream_id"]) -> None: ...

global___PublishBuildToolEventStreamResponse = PublishBuildToolEventStreamResponse

@typing_extensions.final
class OrderedBuildEvent(google.protobuf.message.Message):
    """Build event with contextual information about the stream it belongs to and
    its position in that stream.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STREAM_ID_FIELD_NUMBER: builtins.int
    SEQUENCE_NUMBER_FIELD_NUMBER: builtins.int
    EVENT_FIELD_NUMBER: builtins.int
    @property
    def stream_id(self) -> buildgrid._protos.google.devtools.build.v1.build_events_pb2.StreamId:
        """Which build event stream this event belongs to."""
    sequence_number: builtins.int
    """The position of this event in the stream. The sequence numbers for a build
    event stream should be a sequence of consecutive natural numbers starting
    from one. (1, 2, 3, ...)
    """
    @property
    def event(self) -> buildgrid._protos.google.devtools.build.v1.build_events_pb2.BuildEvent:
        """The actual event."""
    def __init__(
        self,
        *,
        stream_id: buildgrid._protos.google.devtools.build.v1.build_events_pb2.StreamId | None = ...,
        sequence_number: builtins.int = ...,
        event: buildgrid._protos.google.devtools.build.v1.build_events_pb2.BuildEvent | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["event", b"event", "stream_id", b"stream_id"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["event", b"event", "sequence_number", b"sequence_number", "stream_id", b"stream_id"]) -> None: ...

global___OrderedBuildEvent = OrderedBuildEvent

@typing_extensions.final
class PublishBuildToolEventStreamRequest(google.protobuf.message.Message):
    """Streaming request message for PublishBuildToolEventStream."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ORDERED_BUILD_EVENT_FIELD_NUMBER: builtins.int
    NOTIFICATION_KEYWORDS_FIELD_NUMBER: builtins.int
    PROJECT_ID_FIELD_NUMBER: builtins.int
    @property
    def ordered_build_event(self) -> global___OrderedBuildEvent:
        """Required. The build event with position info.
        New publishing clients should use this field rather than the 3 above.
        """
    @property
    def notification_keywords(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """The keywords to be attached to the notification which notifies the start
        of a new build event stream. BES only reads this field when sequence_number
        or ordered_build_event.sequence_number is 1 in this message. If this field
        is empty, BES will not publish notification messages for this stream.
        """
    project_id: builtins.str
    """Required. The project this build is associated with.
    This should match the project used for the initial call to
    PublishLifecycleEvent (containing a BuildEnqueued message).
    """
    def __init__(
        self,
        *,
        ordered_build_event: global___OrderedBuildEvent | None = ...,
        notification_keywords: collections.abc.Iterable[builtins.str] | None = ...,
        project_id: builtins.str = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["ordered_build_event", b"ordered_build_event"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["notification_keywords", b"notification_keywords", "ordered_build_event", b"ordered_build_event", "project_id", b"project_id"]) -> None: ...

global___PublishBuildToolEventStreamRequest = PublishBuildToolEventStreamRequest



# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import hashlib
import json
from collections import defaultdict
from dataclasses import dataclass
from itertools import chain, combinations
from typing import Dict, Iterable, List, Optional, Protocol, Set, Tuple

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import Platform
from buildgrid._protos.google.devtools.remoteworkers.v1test2.bots_pb2 import BotSession
from buildgrid.server.exceptions import FailedPreconditionError
from buildgrid.server.logging import buildgrid_logger

CAPABILITIES_WARNING_THRESHOLD = 10
LOGGER = buildgrid_logger(__name__)


def hash_from_dict(dictionary: Dict[str, List[str]]) -> str:
    """Get the hash represntation of a dictionary"""
    return hashlib.sha1(json.dumps(dictionary, sort_keys=True).encode()).hexdigest()


class PropertySet(Protocol):
    def execution_properties(self, platform: Platform) -> Tuple[str, Dict[str, List[str]]]:
        """
        Parses a platform value and returns the match properties used for scheduling.
        Returns a label which can be used for applying metrics.
        """

    def worker_properties(self, bot_session: BotSession) -> List[Dict[str, List[str]]]:
        """
        Find all the valid property combinations which can be used to assign work to a bot.
        """


class DynamicPropertySet:
    def __init__(
        self,
        *,
        unique_property_keys: Set[str],
        match_property_keys: Set[str],
        wildcard_property_keys: Set[str],
        label_key: Optional[str] = None,
    ) -> None:
        if unregistered_unique_keys := (unique_property_keys - match_property_keys) - wildcard_property_keys:
            raise ValueError(f"Unique keys configured which are not match or wildcards: {unregistered_unique_keys}")

        if label_key and label_key not in match_property_keys and label_key not in wildcard_property_keys:
            raise ValueError(f"Label key is not registered as a match or wildcard key: {label_key}")

        self.unique_property_keys = set(unique_property_keys)
        self.match_property_keys = set(match_property_keys)
        self.wildcard_property_keys = set(wildcard_property_keys)
        self.all_property_keys = match_property_keys | wildcard_property_keys
        self.label_key = label_key

    def execution_properties(self, platform: Platform) -> Tuple[str, Dict[str, List[str]]]:
        properties: Dict[str, Set[str]] = defaultdict(set)
        for platform_property in platform.properties:
            properties[platform_property.name].add(platform_property.value)

        label = "unknown"
        if self.label_key in properties:
            label = sorted(properties[self.label_key])[0]

        for name, values in properties.items():
            if name not in self.all_property_keys:
                raise FailedPreconditionError(
                    f"Unregistered platform property [{name}={values}]."
                    f" Known properties are: [{self.all_property_keys}]"
                )
            if name in self.unique_property_keys and len(values) > 1:
                raise FailedPreconditionError(
                    f"Unique platform property [{name}] can only be set once. Got: [{values}]"
                )

        result = {k: sorted(v) for k, v in properties.items() if k in self.match_property_keys}
        return label, result

    def worker_properties(self, bot_session: BotSession) -> List[Dict[str, List[str]]]:
        properties = bot_properties(bot_session)
        properties = {k: v for k, v in properties.items() if k in self.match_property_keys}
        return partial_bot_properties(properties)


Properties = Set[Tuple[str, str]]


@dataclass
class PropertyLabel:
    label: str
    properties: Properties


class StaticPropertySet:
    def __init__(
        self,
        *,
        property_labels: List[PropertyLabel],
        wildcard_property_keys: Set[str],
    ) -> None:
        self.property_labels = property_labels
        self.wildcard_property_keys = wildcard_property_keys

    def execution_properties(self, platform: Platform) -> Tuple[str, Dict[str, List[str]]]:
        execute_properties = {
            (platform_property.name, platform_property.value)
            for platform_property in platform.properties
            if platform_property.name not in self.wildcard_property_keys
        }

        for property_label in self.property_labels:
            if len(execute_properties - property_label.properties) == 0:
                return property_label.label, merge_property_pairs(property_label.properties)

        raise FailedPreconditionError(f"Could not find property set for {execute_properties}")

    def worker_properties(self, bot_session: BotSession) -> List[Dict[str, List[str]]]:
        bots_properties = bot_properties(bot_session)
        property_pairs = {
            (key, value)
            for key, values in bots_properties.items()
            for value in values
            if key not in self.wildcard_property_keys
        }

        property_sets = []
        for property_set in self.property_labels:
            if len(property_set.properties - property_pairs) == 0:
                property_sets.append(merge_property_pairs(property_set.properties))

        if len(property_sets) == 0:
            raise FailedPreconditionError(f"Could not find property set for {bots_properties}")
        return [{k: sorted(v) for k, v in props.items()} for props in property_sets]


def bot_properties(bot_session: BotSession) -> Dict[str, Set[str]]:
    worker_capabilities: Dict[str, Set[str]] = {}
    if bot_session.worker.devices:
        # According to the spec:
        #   "The first device in the worker is the "primary device" -
        #   that is, the device running a bot and which is
        #   responsible for actually executing commands."
        primary_device = bot_session.worker.devices[0]

        for device_property in primary_device.properties:
            if device_property.key not in worker_capabilities:
                worker_capabilities[device_property.key] = set()
            worker_capabilities[device_property.key].add(device_property.value)
    return worker_capabilities


def partial_bot_properties(properties: Dict[str, Set[str]]) -> List[Dict[str, List[str]]]:
    property_pairs = flatten_properties(properties)

    if len(property_pairs) > CAPABILITIES_WARNING_THRESHOLD:
        LOGGER.warning(
            "A worker with a large capabilities dictionary has been connected. "
            f"Processing its capabilities may take a while. Capabilities: {property_pairs}"
        )

    # Using the itertools powerset recipe, construct the powerset of the tuples
    powerset = chain.from_iterable(combinations(property_pairs, r) for r in range(len(property_pairs) + 1))
    return list(map(merge_property_pairs, powerset))


def flatten_properties(properties: Dict[str, Set[str]]) -> List[Tuple[str, str]]:
    return [(name, value) for name in sorted(properties) for value in sorted(properties[name])]


def merge_property_pairs(property_pairs: Iterable[Tuple[str, str]]) -> Dict[str, List[str]]:
    properties: Dict[str, List[str]] = {}
    for name, value in property_pairs:
        properties.setdefault(name, []).append(value)
    return {k: sorted(v) for k, v in properties.items()}



# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from dataclasses import dataclass
from typing import Optional

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import Digest, DigestFunction
from buildgrid.server.settings import HASH, HASH_LENGTH


@dataclass(frozen=True)
class HashableDigest:
    hash: str
    size_bytes: int

    def to_digest(self) -> Digest:
        return Digest(hash=self.hash, size_bytes=self.size_bytes)


def get_hash_type() -> "DigestFunction.Value.ValueType":
    """Returns the hash type."""
    hash_name = HASH().name
    if hash_name == "sha256":
        return DigestFunction.SHA256
    return DigestFunction.UNKNOWN


def create_digest(bytes_to_digest: bytes) -> Digest:
    """Computes the :obj:`Digest` of a piece of data.

    The :obj:`Digest` of a data is a function of its hash **and** size.

    Args:
        bytes_to_digest (bytes): byte data to digest.

    Returns:
        :obj:`Digest`: The :obj:`Digest` for the given byte data.
    """
    return Digest(hash=HASH(bytes_to_digest).hexdigest(), size_bytes=len(bytes_to_digest))


def parse_digest(digest_string: str) -> Optional[Digest]:
    """Creates a :obj:`Digest` from a digest string.

    A digest string should alway be: ``{hash}/{size_bytes}``.

    Args:
        digest_string (str): the digest string.

    Returns:
        :obj:`Digest`: The :obj:`Digest` read from the string or None if
            `digest_string` is not a valid digest string.
    """
    digest_hash, digest_size = digest_string.split("/")

    if len(digest_hash) == HASH_LENGTH and digest_size.isdigit():
        return Digest(hash=digest_hash, size_bytes=int(digest_size))

    return None


def validate_digest_data(digest: Digest, data: bytes) -> bool:
    """Validate that the given digest corresponds to the given data."""
    return len(data) == digest.size_bytes and HASH(data).hexdigest() == digest.hash



from typing import Any, Callable, Dict, TypeVar, overload

T = TypeVar("T")


class _Unset:
    pass


@overload
def map_key(conf: Dict[str, Any], key: str) -> Any: ...


@overload
def map_key(conf: Dict[str, Any], key: str, *, default: T) -> T: ...


@overload
def map_key(conf: Dict[str, Any], key: str, *, decoder: Callable[[Any], T]) -> T: ...


@overload
def map_key(conf: Dict[str, Any], key: str, *, decoder: Callable[[Any], T], default: T) -> T: ...


def map_key(
    conf: Dict[str, Any], key: str, *, decoder: Callable[[Any], Any] = lambda v: v, default: Any = _Unset()
) -> Any:
    if key in conf:
        return decoder(conf[key])
    if isinstance(default, _Unset):
        raise ValueError(f"Unset config value for {key}")
    return default



# Copyright (C) 2021 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Optional

from aiohttp_middlewares.annotations import UrlCollection
from aiohttp_middlewares.cors import ACCESS_CONTROL_ALLOW_CREDENTIALS, ACCESS_CONTROL_ALLOW_ORIGIN

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import ActionResult
from buildgrid.server.settings import (
    BROWSER_BLOB_CACHE_MAX_LENGTH,
    BROWSER_BLOB_CACHE_TTL,
    BROWSER_RESULT_CACHE_MAX_LENGTH,
    BROWSER_RESULT_CACHE_TTL,
)
from buildgrid.server.utils.async_lru_cache import LruCache

TARBALL_DIRECTORY_PREFIX = "bgd-browser-tarball-"


def get_cors_headers(origin: Optional[str], origins: UrlCollection, allow_all: bool) -> Dict[str, str]:
    headers = {}
    if origin:
        headers[ACCESS_CONTROL_ALLOW_CREDENTIALS] = "true"
        if origin in origins:
            headers[ACCESS_CONTROL_ALLOW_ORIGIN] = origin
        elif allow_all:
            headers[ACCESS_CONTROL_ALLOW_ORIGIN] = "*"
    return headers


class ResponseCache:
    def __init__(self) -> None:
        self._action_results: LruCache[ActionResult] = LruCache(BROWSER_RESULT_CACHE_MAX_LENGTH)
        self._blobs: LruCache[bytes] = LruCache(BROWSER_BLOB_CACHE_MAX_LENGTH)

    async def store_action_result(self, key: str, result: ActionResult) -> None:
        await self._action_results.set(key, result, BROWSER_RESULT_CACHE_TTL)

    async def get_action_result(self, key: str) -> Optional[ActionResult]:
        return await self._action_results.get(key)

    async def store_blob(self, key: str, blob: bytes) -> None:
        await self._blobs.set(key, blob, BROWSER_BLOB_CACHE_TTL)

    async def get_blob(self, key: str) -> Optional[bytes]:
        return await self._blobs.get(key)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: build/buildbox/execution_stats.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n$build/buildbox/execution_stats.proto\x12\x0e\x62uild.buildbox\x1a\x1egoogle/protobuf/duration.proto\"\xc5\x02\n\x13\x45xecutionStatistics\x12P\n\x0e\x63ommand_rusage\x18\x01 \x01(\x0b\x32\x38.build.buildbox.ExecutionStatistics.ProcessResourceUsage\x1a\xdb\x01\n\x14ProcessResourceUsage\x12(\n\x05utime\x18\x01 \x01(\x0b\x32\x19.google.protobuf.Duration\x12(\n\x05stime\x18\x02 \x01(\x0b\x32\x19.google.protobuf.Duration\x12\x0e\n\x06maxrss\x18\x03 \x01(\x03\x12\x0e\n\x06minflt\x18\x04 \x01(\x03\x12\x0e\n\x06majflt\x18\x05 \x01(\x03\x12\x0f\n\x07inblock\x18\x06 \x01(\x03\x12\x0f\n\x07oublock\x18\x07 \x01(\x03\x12\r\n\x05nvcsw\x18\x08 \x01(\x03\x12\x0e\n\x06nivcsw\x18\t \x01(\x03\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'buildgrid._protos.build.buildbox.execution_stats_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_EXECUTIONSTATISTICS']._serialized_start=89
  _globals['_EXECUTIONSTATISTICS']._serialized_end=414
  _globals['_EXECUTIONSTATISTICS_PROCESSRESOURCEUSAGE']._serialized_start=195
  _globals['_EXECUTIONSTATISTICS_PROCESSRESOURCEUSAGE']._serialized_end=414
# @@protoc_insertion_point(module_scope)



# Copyright (C) 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add index on queued_timestamp in jobs table

Revision ID: 083f52d97bcb
Revises: 69ceb5938c06
Create Date: 2020-06-01 22:05:44.274091

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "083f52d97bcb"
down_revision = "69ceb5938c06"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_index(op.f("ix_jobs_queued_timestamp"), "jobs", ["queued_timestamp"], unique=False)


def downgrade() -> None:
    op.drop_index(op.f("ix_jobs_queued_timestamp"), table_name="jobs")



# Copyright (C) 2019 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple
from typing import TYPE_CHECKING, Any, List, Optional, Tuple, Union
from urllib.parse import urlparse

import grpc
from grpc import aio

from buildgrid._protos.build.bazel.remote.execution.v2 import remote_execution_pb2
from buildgrid.server.client.auth_token_loader import AuthTokenLoader
from buildgrid.server.client.authentication import AuthMetadataClientInterceptorBase, load_tls_channel_credentials
from buildgrid.server.client.interceptors import (
    AsyncStreamStreamInterceptor,
    AsyncStreamUnaryInterceptor,
    AsyncUnaryStreamInterceptor,
    AsyncUnaryUnaryInterceptor,
    SyncStreamStreamInterceptor,
    SyncStreamUnaryInterceptor,
    SyncUnaryStreamInterceptor,
    SyncUnaryUnaryInterceptor,
)
from buildgrid.server.exceptions import InvalidArgumentError
from buildgrid.server.settings import (
    INSECURE_URI_SCHEMES,
    REQUEST_METADATA_HEADER_NAME,
    REQUEST_METADATA_TOOL_NAME,
    REQUEST_METADATA_TOOL_VERSION,
    SECURE_URI_SCHEMES,
)


def setup_channel(
    remote_url: str,
    auth_token: Optional[str] = None,
    auth_token_refresh_seconds: Optional[int] = None,
    client_key: Optional[str] = None,
    client_cert: Optional[str] = None,
    server_cert: Optional[str] = None,
    action_id: Optional[str] = None,
    tool_invocation_id: Optional[str] = None,
    correlated_invocations_id: Optional[str] = None,
    asynchronous: bool = False,
    timeout: Optional[float] = None,
) -> Tuple[grpc.Channel, Tuple[Optional[str], ...]]:
    """Creates a new gRPC client communication chanel.

    If `remote_url` does not point to a socket and does not specify a
    port number, defaults 50051.

    Args:
        remote_url (str): URL for the remote, including protocol and,
            if not a Unix domain socket, a port.
        auth_token (str): Authorization token file path.
        auth_token_refresh_seconds(int): Time in seconds to read the authorization token again from file
        server_cert(str): TLS certificate chain file path.
        client_key (str): TLS root certificate file path.
        client_cert (str): TLS private key file path.
        action_id (str): Action identifier to which the request belongs to.
        tool_invocation_id (str): Identifier for a related group of Actions.
        correlated_invocations_id (str): Identifier that ties invocations together.
        timeout (float): Request timeout in seconds.

    Returns:
        Channel: Client Channel to be used in order to access the server
            at `remote_url`.

    Raises:
        InvalidArgumentError: On any input parsing error.
    """
    url = urlparse(remote_url)

    url_is_socket = url.scheme == "unix"
    if url_is_socket:
        remote = remote_url
    else:
        remote = f"{url.hostname}:{url.port or 50051}"

    details: Tuple[Optional[str], Optional[str], Optional[str]] = None, None, None
    credentials_provided = any((server_cert, client_cert, client_key))
    auth_token_loader: Optional[AuthTokenLoader] = None
    if auth_token:
        auth_token_loader = AuthTokenLoader(auth_token, auth_token_refresh_seconds)

    if asynchronous:
        async_interceptors = _create_async_interceptors(
            auth_token_loader=auth_token_loader,
            action_id=action_id,
            tool_invocation_id=tool_invocation_id,
            correlated_invocations_id=correlated_invocations_id,
            timeout=timeout,
        )

        if url.scheme in INSECURE_URI_SCHEMES or (url_is_socket and not credentials_provided):
            async_channel = aio.insecure_channel(remote, interceptors=async_interceptors)
        elif url.scheme in SECURE_URI_SCHEMES or (url_is_socket and credentials_provided):
            credentials, details = load_tls_channel_credentials(client_key, client_cert, server_cert)
            if not credentials:
                raise InvalidArgumentError("Given TLS details (or defaults) could be loaded")
            async_channel = aio.secure_channel(remote, credentials, interceptors=async_interceptors)
        else:
            raise InvalidArgumentError("Given remote does not specify a protocol")

        # TODO use overloads to make this return an async channel when asynchronous == True
        return async_channel, details  # type: ignore[return-value]

    else:
        sync_interceptors = _create_sync_interceptors(
            auth_token_loader=auth_token_loader,
            action_id=action_id,
            tool_invocation_id=tool_invocation_id,
            correlated_invocations_id=correlated_invocations_id,
            timeout=timeout,
        )

        if url.scheme in INSECURE_URI_SCHEMES or (url_is_socket and not credentials_provided):
            sync_channel = grpc.insecure_channel(remote)
        elif url.scheme in SECURE_URI_SCHEMES or (url_is_socket and credentials_provided):
            credentials, details = load_tls_channel_credentials(client_key, client_cert, server_cert)
            if not credentials:
                raise InvalidArgumentError("Given TLS details (or defaults) could be loaded")
            sync_channel = grpc.secure_channel(remote, credentials)
        else:
            raise InvalidArgumentError("Given remote does not specify a protocol")

        for interceptor in sync_interceptors:
            sync_channel = grpc.intercept_channel(sync_channel, interceptor)

        return sync_channel, details


class RequestMetadataInterceptorBase:
    def __init__(
        self,
        action_id: Optional[str] = None,
        tool_invocation_id: Optional[str] = None,
        correlated_invocations_id: Optional[str] = None,
    ) -> None:
        """Appends optional `RequestMetadata` header values to each call.

        Args:
            action_id (str): Action identifier to which the request belongs to.
            tool_invocation_id (str): Identifier for a related group of Actions.
            correlated_invocations_id (str): Identifier that ties invocations together.
        """
        self._action_id = action_id
        self._tool_invocation_id = tool_invocation_id
        self._correlated_invocations_id = correlated_invocations_id

        self.__header_field_name = REQUEST_METADATA_HEADER_NAME
        self.__header_field_value = self._request_metadata()

    def _request_metadata(self) -> bytes:
        """Creates a serialized RequestMetadata entry to attach to a gRPC
        call header. Arguments should be of type str or None.
        """
        request_metadata = remote_execution_pb2.RequestMetadata()
        request_metadata.tool_details.tool_name = REQUEST_METADATA_TOOL_NAME
        request_metadata.tool_details.tool_version = REQUEST_METADATA_TOOL_VERSION

        if self._action_id:
            request_metadata.action_id = self._action_id
        if self._tool_invocation_id:
            request_metadata.tool_invocation_id = self._tool_invocation_id
        if self._correlated_invocations_id:
            request_metadata.correlated_invocations_id = self._correlated_invocations_id

        return request_metadata.SerializeToString()

    def amend_call_details(  # type: ignore[no-untyped-def] # wait for client lib updates here
        self, client_call_details, grpc_call_details_class: Any
    ):
        if client_call_details.metadata is not None:
            new_metadata = list(client_call_details.metadata)
        else:
            new_metadata = []

        new_metadata.append((self.__header_field_name, self.__header_field_value))

        class _ClientCallDetails(
            namedtuple(
                "_ClientCallDetails",
                (
                    "method",
                    "timeout",
                    "credentials",
                    "metadata",
                    "wait_for_ready",
                ),
            ),
            grpc_call_details_class,  # type: ignore
        ):
            pass

        return _ClientCallDetails(
            client_call_details.method,
            client_call_details.timeout,
            client_call_details.credentials,
            new_metadata,
            client_call_details.wait_for_ready,
        )


class TimeoutInterceptorBase:
    def __init__(self, timeout: float) -> None:
        """Applies a request timeout to each call.

        Args:
            timeout (float): Request timeout in seconds.
        """
        self._timeout = timeout

    def amend_call_details(  # type: ignore[no-untyped-def] # wait for client lib updates here
        self, client_call_details, grpc_call_details_class: Any
    ):
        # If there are multiple timeouts, apply the shorter timeout (earliest deadline wins)
        if client_call_details.timeout is not None:
            new_timeout = min(self._timeout, client_call_details.timeout)
        else:
            new_timeout = self._timeout

        class _ClientCallDetails(
            namedtuple(
                "_ClientCallDetails",
                (
                    "method",
                    "timeout",
                    "credentials",
                    "metadata",
                    "wait_for_ready",
                ),
            ),
            grpc_call_details_class,  # type: ignore
        ):
            pass

        return _ClientCallDetails(
            client_call_details.method,
            new_timeout,
            client_call_details.credentials,
            client_call_details.metadata,
            client_call_details.wait_for_ready,
        )


if TYPE_CHECKING:
    SyncInterceptorsList = List[
        Union[
            grpc.UnaryUnaryClientInterceptor[Any, Any],
            grpc.UnaryStreamClientInterceptor[Any, Any],
            grpc.StreamUnaryClientInterceptor[Any, Any],
            grpc.StreamStreamClientInterceptor[Any, Any],
        ]
    ]


def _create_sync_interceptors(
    auth_token_loader: Optional[AuthTokenLoader] = None,
    action_id: Optional[str] = None,
    tool_invocation_id: Optional[str] = None,
    correlated_invocations_id: Optional[str] = None,
    timeout: Optional[float] = None,
) -> "SyncInterceptorsList":
    interceptors: "SyncInterceptorsList" = []

    request_metadata_interceptor = RequestMetadataInterceptorBase(
        action_id=action_id,
        tool_invocation_id=tool_invocation_id,
        correlated_invocations_id=correlated_invocations_id,
    )

    interceptors += [
        SyncUnaryUnaryInterceptor(request_metadata_interceptor),
        SyncUnaryStreamInterceptor(request_metadata_interceptor),
        SyncStreamUnaryInterceptor(request_metadata_interceptor),
        SyncStreamStreamInterceptor(request_metadata_interceptor),
    ]

    if auth_token_loader is not None:
        auth_metadata_client_interceptor = AuthMetadataClientInterceptorBase(auth_token_loader=auth_token_loader)
        interceptors += [
            SyncUnaryUnaryInterceptor(auth_metadata_client_interceptor),
            SyncUnaryStreamInterceptor(auth_metadata_client_interceptor),
            SyncStreamUnaryInterceptor(auth_metadata_client_interceptor),
            SyncStreamStreamInterceptor(auth_metadata_client_interceptor),
        ]

    if timeout is not None:
        timeout_interceptor_base = TimeoutInterceptorBase(timeout)
        interceptors += [
            SyncUnaryUnaryInterceptor(timeout_interceptor_base),
            SyncUnaryStreamInterceptor(timeout_interceptor_base),
            SyncStreamUnaryInterceptor(timeout_interceptor_base),
            SyncStreamStreamInterceptor(timeout_interceptor_base),
        ]

    return interceptors


def _create_async_interceptors(
    auth_token_loader: Optional[AuthTokenLoader] = None,
    action_id: Optional[str] = None,
    tool_invocation_id: Optional[str] = None,
    correlated_invocations_id: Optional[str] = None,
    timeout: Optional[float] = None,
) -> List[aio.ClientInterceptor]:
    # FIXME Types not happy...  "list" has incompatible type "..."; expected "_PartialStubMustCastOrIgnore"
    interceptors: List[Any] = []

    request_metadata_interceptor = RequestMetadataInterceptorBase(
        action_id=action_id,
        tool_invocation_id=tool_invocation_id,
        correlated_invocations_id=correlated_invocations_id,
    )

    interceptors += [
        AsyncUnaryUnaryInterceptor(request_metadata_interceptor),
        AsyncUnaryStreamInterceptor(request_metadata_interceptor),
        AsyncStreamUnaryInterceptor(request_metadata_interceptor),
        AsyncStreamStreamInterceptor(request_metadata_interceptor),
    ]

    if auth_token_loader is not None:
        auth_metadata_client_interceptor = AuthMetadataClientInterceptorBase(auth_token_loader=auth_token_loader)
        interceptors += [
            AsyncUnaryUnaryInterceptor(auth_metadata_client_interceptor),
            AsyncUnaryStreamInterceptor(auth_metadata_client_interceptor),
            AsyncStreamUnaryInterceptor(auth_metadata_client_interceptor),
            AsyncStreamStreamInterceptor(auth_metadata_client_interceptor),
        ]

    if timeout is not None:
        timeout_interceptor_base = TimeoutInterceptorBase(timeout)
        interceptors += [
            AsyncUnaryUnaryInterceptor(timeout_interceptor_base),
            AsyncUnaryStreamInterceptor(timeout_interceptor_base),
            AsyncStreamUnaryInterceptor(timeout_interceptor_base),
            AsyncStreamStreamInterceptor(timeout_interceptor_base),
        ]

    return interceptors



# Copyright (C) 2021 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License' is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import sys
from typing import Optional

import click
from aiohttp import web
from aiohttp_middlewares.annotations import UrlCollection

from buildgrid.server.browser.app import create_app
from buildgrid.server.browser.utils import ResponseCache
from buildgrid.server.client.channel import setup_channel
from buildgrid.server.exceptions import InvalidArgumentError

from ..cli import Context, pass_context, setup_logging

pass_cache = click.make_pass_decorator(ResponseCache, ensure=True)


@click.group(name="browser-backend", short_help="bgd-browser backend API.")
@click.option(
    "--remote",
    type=click.STRING,
    default="http://localhost:50051",
    show_default=True,
    help="Remote server's URL (port defaults to 50051 if not specified).",
)
@click.option(
    "--remote-cas",
    type=click.STRING,
    default=None,
    show_default=False,
    help="Remote CAS server's URL, defaults to `--remote` if unspecified.",
)
@click.option(
    "--remote-operations",
    type=click.STRING,
    default=None,
    show_default=False,
    help="Remote Operations server's URL, defaults to `--remote` if unspecified.",
)
@click.option(
    "--remote-cache",
    type=click.STRING,
    default=None,
    show_default=False,
    help="Remote ActionCache server's URL, defaults to `--remote` if unspecified.",
)
@click.option(
    "--remote-logstream",
    type=click.STRING,
    default=None,
    show_default=False,
    help="Remote LogStream server's URL, defaults to `--remote` if unspecified.",
)
@click.option(
    "--auth-token",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Authorization token for the remote.",
)
@click.option(
    "--client-key",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Private client key for TLS (PEM-encoded).",
)
@click.option(
    "--client-cert",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Public client certificate for TLS (PEM-encoded).",
)
@click.option(
    "--server-cert",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Public server certificate for TLS (PEM-encoded).",
)
@click.option(
    "--instance-name", type=click.STRING, default="", show_default=True, help="Targeted BuildGrid instance name."
)
@pass_context
def cli(
    context: Context,
    remote: str,
    remote_cas: str,
    remote_operations: str,
    remote_cache: str,
    remote_logstream: str,
    instance_name: str,
    auth_token: Optional[str],
    client_key: Optional[str],
    client_cert: Optional[str],
    server_cert: Optional[str],
) -> None:
    try:
        context.channel, _ = setup_channel(
            remote,
            auth_token=auth_token,
            client_key=client_key,
            client_cert=client_cert,
            server_cert=server_cert,
            asynchronous=True,
        )

        if remote_cas and remote_cas != remote:
            context.cas_channel, _ = setup_channel(
                remote_cas,
                auth_token=auth_token,
                client_key=client_key,
                client_cert=client_cert,
                server_cert=server_cert,
                asynchronous=True,
            )
        else:
            context.cas_channel = context.channel

        if remote_operations and remote_operations != remote:
            context.operations_channel, _ = setup_channel(
                remote_operations,
                auth_token=auth_token,
                client_key=client_key,
                client_cert=client_cert,
                server_cert=server_cert,
                asynchronous=True,
            )
        else:
            context.operations_channel = context.channel

        if remote_cache and remote_cache != remote:
            context.cache_channel, _ = setup_channel(
                remote_cache,
                auth_token=auth_token,
                client_key=client_key,
                client_cert=client_cert,
                server_cert=server_cert,
                asynchronous=True,
            )
        else:
            context.cache_channel = context.channel

        if remote_logstream and remote_logstream != remote:
            context.logstream_channel, _ = setup_channel(
                remote_logstream,
                auth_token=auth_token,
                client_key=client_key,
                client_cert=client_cert,
                server_cert=server_cert,
                asynchronous=True,
            )
        else:
            context.logstream_channel = context.channel

    except InvalidArgumentError as e:
        click.echo(f"Error: {e}.", err=True)
        sys.exit(-1)

    context.instance_name = instance_name


@cli.command("serve", short_help="Serve the bgd-browser backend.")
@click.option("--port", "-p", type=click.INT, default=8083, show_default=True, help="Port to serve the API on")
@click.option(
    "--cors-allowed-origin",
    "-o",
    multiple=True,
    default=[],
    show_default=True,
    help="URLs to allow CORS requests from.",
)
@click.option(
    "--allow-cancelling-operations",
    default=False,
    show_default=True,
    help="Forward DELETE requests to the 'operations/' endpoint as CancelOperation() RPCs.",
)
@click.option("-v", "--verbose", count=True, help="Increase log verbosity level.")
@click.option(
    "--static-path", "-s", type=click.STRING, default=None, help="Optional path to a directory to serve at /"
)
@click.option(
    "--tarball-directory",
    "-t",
    type=click.STRING,
    default=None,
    help="Path to a directory to use when constructing tarballs",
)
@pass_cache
@pass_context
def serve(
    context: Context,
    cache: ResponseCache,
    port: int,
    verbose: int,
    cors_allowed_origin: UrlCollection,
    allow_cancelling_operations: bool,
    static_path: Optional[str],
    tarball_directory: Optional[str],
) -> None:
    setup_logging(verbosity=verbose)
    app = create_app(
        context,
        cache,
        cors_allowed_origin,
        allow_cancelling_operations=allow_cancelling_operations,
        static_path=static_path,
        tarball_dir=tarball_directory,
    )

    # TODO Refactor the async channel setup to not happen before we enter the web app.
    # The latest version of aiohttp will make a new event loop on startup which currently
    # does not match the loop used when creating the GRPC channel objects.
    #
    # This is caused by https://github.com/aio-libs/aiohttp/pull/5572/files
    #
    # For right now we can mitigate the issue by having the current event loop be forced in
    # the aiohttp.web.Application.
    web.run_app(app, port=port, loop=asyncio.get_event_loop())



# Copyright (C) 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add client_identities table

Revision ID: b57c30a687fa
Revises: 53e8b8b5bed6
Create Date: 2023-04-11 14:22:08.511246

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "b57c30a687fa"
down_revision = "53e8b8b5bed6"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "client_identities",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("instance", sa.String(), nullable=False),
        sa.Column("workflow", sa.String(), nullable=False),
        sa.Column("actor", sa.String(), nullable=False),
        sa.Column("subject", sa.String(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("instance", "workflow", "actor", "subject"),
    )

    # Adjusted by using batch_op and supplying a constraint name to be compatible with SQLight
    with op.batch_alter_table("operations") as batch_op:
        batch_op.add_column(sa.Column("client_identity_id", sa.Integer(), nullable=True))
        batch_op.create_foreign_key(
            "operation__client_identities_id", "client_identities", ["client_identity_id"], ["id"]
        )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###

    # Adjusted by using batch_op and supplying a constraint name to be compatible with SQLight
    with op.batch_alter_table("operations") as batch_op:
        batch_op.drop_constraint("operation__client_identities_id", type_="foreignkey")
        batch_op.drop_column("client_identity_id")
    op.drop_table("client_identities")
    # ### end Alembic commands ###



# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
CAS services
==================

Implements the Content Addressable Storage API and ByteStream API.
"""

import itertools
import re
from typing import Any, Dict, Iterator, Tuple, cast

import grpc

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import DESCRIPTOR as RE_DESCRIPTOR
from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import (
    BatchReadBlobsRequest,
    BatchReadBlobsResponse,
    BatchUpdateBlobsRequest,
    BatchUpdateBlobsResponse,
    Digest,
    FindMissingBlobsRequest,
    FindMissingBlobsResponse,
    GetTreeRequest,
    GetTreeResponse,
)
from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2_grpc import (
    ContentAddressableStorageServicer,
    add_ContentAddressableStorageServicer_to_server,
)
from buildgrid._protos.google.bytestream import bytestream_pb2, bytestream_pb2_grpc
from buildgrid._protos.google.bytestream.bytestream_pb2 import (
    QueryWriteStatusRequest,
    QueryWriteStatusResponse,
    ReadRequest,
    ReadResponse,
    WriteRequest,
    WriteResponse,
)
from buildgrid._protos.google.rpc import code_pb2, status_pb2
from buildgrid.server.cas.instance import (
    EMPTY_BLOB,
    EMPTY_BLOB_DIGEST,
    ByteStreamInstance,
    ContentAddressableStorageInstance,
)
from buildgrid.server.decorators import rpc
from buildgrid.server.enums import ByteStreamResourceType
from buildgrid.server.exceptions import InvalidArgumentError
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.servicer import InstancedServicer
from buildgrid.server.settings import HASH_LENGTH

LOGGER = buildgrid_logger(__name__)


def _printable_batch_update_blobs_request(request: BatchUpdateBlobsRequest) -> Dict[str, Any]:
    # Log the digests but not the data
    return {
        "instance_name": request.instance_name,
        "digests": [r.digest for r in request.requests],
    }


class ContentAddressableStorageService(
    ContentAddressableStorageServicer, InstancedServicer[ContentAddressableStorageInstance]
):
    SERVICE_NAME = "ContentAddressableStorage"
    REGISTER_METHOD = add_ContentAddressableStorageServicer_to_server
    FULL_NAME = RE_DESCRIPTOR.services_by_name[SERVICE_NAME].full_name

    @rpc(instance_getter=lambda r: cast(str, r.instance_name))
    def FindMissingBlobs(
        self, request: FindMissingBlobsRequest, context: grpc.ServicerContext
    ) -> FindMissingBlobsResponse:
        # No need to find the empty blob in the cas because the empty blob cannot be missing
        digests_to_find = [digest for digest in request.blob_digests if digest != EMPTY_BLOB_DIGEST]
        return self.get_instance(request.instance_name).find_missing_blobs(digests_to_find)

    @rpc(instance_getter=lambda r: cast(str, r.instance_name), request_formatter=_printable_batch_update_blobs_request)
    def BatchUpdateBlobs(
        self, request: BatchUpdateBlobsRequest, context: grpc.ServicerContext
    ) -> BatchUpdateBlobsResponse:
        return self.get_instance(request.instance_name).batch_update_blobs(request.requests)

    @rpc(instance_getter=lambda r: cast(str, r.instance_name))
    def BatchReadBlobs(self, request: BatchReadBlobsRequest, context: grpc.ServicerContext) -> BatchReadBlobsResponse:
        # No need to actually read the empty blob in the cas as it is always present
        digests_to_read = [digest for digest in request.digests if digest != EMPTY_BLOB_DIGEST]
        empty_digest_count = len(request.digests) - len(digests_to_read)

        instance = self.get_instance(request.instance_name)
        response = instance.batch_read_blobs(digests_to_read)

        # Append the empty blobs to the response
        for _ in range(empty_digest_count):
            response_proto = response.responses.add()
            response_proto.data = EMPTY_BLOB
            response_proto.digest.CopyFrom(EMPTY_BLOB_DIGEST)
            status_code = code_pb2.OK
            response_proto.status.CopyFrom(status_pb2.Status(code=status_code))

        return response

    @rpc(instance_getter=lambda r: cast(str, r.instance_name))
    def GetTree(self, request: GetTreeRequest, context: grpc.ServicerContext) -> Iterator[GetTreeResponse]:
        yield from self.get_instance(request.instance_name).get_tree(request)


class ResourceNameRegex:
    # CAS read name format: "{instance_name}/blobs/{hash}/{size}"
    READ = "^(.*?)/?(blobs/.*/[0-9]*)$"

    # CAS write name format: "{instance_name}/uploads/{uuid}/blobs/{hash}/{size}[optional arbitrary extra content]"
    WRITE = "^(.*?)/?(uploads/.*/blobs/.*/[0-9]*)"


def _parse_resource_name(resource_name: str, regex: str) -> Tuple[str, str, "ByteStreamResourceType"]:
    cas_match = re.match(regex, resource_name)
    if cas_match:
        return cas_match[1], cas_match[2], ByteStreamResourceType.CAS
    else:
        raise InvalidArgumentError(f"Invalid resource name: [{resource_name}]")


def _read_instance_name(resource_name: str) -> str:
    return _parse_resource_name(resource_name, ResourceNameRegex.READ)[0]


def _write_instance_name(resource_name: str) -> str:
    return _parse_resource_name(resource_name, ResourceNameRegex.WRITE)[0]


def _printable_write_request(request: WriteRequest) -> Dict[str, Any]:
    # Log all the fields except `data`:
    return {
        "resource_name": request.resource_name,
        "write_offset": request.write_offset,
        "finish_write": request.finish_write,
    }


class ByteStreamService(bytestream_pb2_grpc.ByteStreamServicer, InstancedServicer[ByteStreamInstance]):
    SERVICE_NAME = "ByteStream"
    REGISTER_METHOD = bytestream_pb2_grpc.add_ByteStreamServicer_to_server
    FULL_NAME = bytestream_pb2.DESCRIPTOR.services_by_name[SERVICE_NAME].full_name

    @rpc(instance_getter=lambda r: _read_instance_name(r.resource_name))
    def Read(self, request: ReadRequest, context: grpc.ServicerContext) -> Iterator[ReadResponse]:
        _, resource_name, resource_type = _parse_resource_name(request.resource_name, ResourceNameRegex.READ)
        if resource_type == ByteStreamResourceType.CAS:
            blob_details = resource_name.split("/")
            if len(blob_details[1]) != HASH_LENGTH:
                raise InvalidArgumentError(f"Invalid digest [{resource_name}]")
            try:
                digest = Digest(hash=blob_details[1], size_bytes=int(blob_details[2]))
            except ValueError:
                raise InvalidArgumentError(f"Invalid digest [{resource_name}]")

            bytes_returned = 0
            expected_bytes = digest.size_bytes - request.read_offset
            if request.read_limit:
                expected_bytes = min(expected_bytes, request.read_limit)

            try:
                if digest.size_bytes == 0:
                    if digest.hash != EMPTY_BLOB_DIGEST.hash:
                        raise InvalidArgumentError(f"Invalid digest [{digest.hash}/{digest.size_bytes}]")
                    yield bytestream_pb2.ReadResponse(data=EMPTY_BLOB)
                    return

                for blob in self.current_instance.read_cas_blob(digest, request.read_offset, request.read_limit):
                    bytes_returned += len(blob.data)
                    yield blob
            finally:
                if bytes_returned != expected_bytes:
                    LOGGER.warning(
                        "Read request exited early.",
                        tags=dict(
                            digest=digest,
                            bytes_returned=bytes_returned,
                            expected_bytes=expected_bytes,
                            read_offset=request.read_offset,
                            read_limit=request.read_limit,
                        ),
                    )
                else:
                    LOGGER.info("Read request completed.", tags=dict(digest=digest))

    @rpc(instance_getter=lambda r: _write_instance_name(r.resource_name), request_formatter=_printable_write_request)
    def Write(self, request_iterator: Iterator[WriteRequest], context: grpc.ServicerContext) -> WriteResponse:
        request = next(request_iterator)
        _, resource_name, resource_type = _parse_resource_name(
            request.resource_name,
            ResourceNameRegex.WRITE,
        )
        if resource_type == ByteStreamResourceType.CAS:
            blob_details = resource_name.split("/")
            _, hash_, size_bytes = blob_details[1], blob_details[3], blob_details[4]
            write_response = self.current_instance.write_cas_blob(
                hash_, size_bytes, itertools.chain([request], request_iterator)
            )
            if write_response.committed_size == int(size_bytes):
                LOGGER.info("Write request completed.", tags=dict(digest=f"{hash_}/{size_bytes}"))
            return write_response
        return bytestream_pb2.WriteResponse()

    @rpc(instance_getter=lambda r: _write_instance_name(r.resource_name))
    def QueryWriteStatus(
        self, request: QueryWriteStatusRequest, context: grpc.ServicerContext
    ) -> QueryWriteStatusResponse:
        context.abort(grpc.StatusCode.UNIMPLEMENTED, "Method not implemented!")



# Copyright (C) 2023 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from time import sleep
from typing import Callable, List, Optional, Tuple, TypeVar

import dns.resolver
import redis
from redis.backoff import EqualJitterBackoff
from redis.exceptions import ConnectionError as RedisConnectionError
from redis.exceptions import TimeoutError as RedisTimeoutError
from redis.sentinel import Sentinel

from buildgrid.server.logging import buildgrid_logger

LOGGER = buildgrid_logger(__name__)

T = TypeVar("T")


class RedisProvider:
    """Provides and manages a Redis connection

    This class manages the connection to a Redis cache.

    The connection can be configured by specifying host/port or by specifying
    a DNS SRV record to use to discover the host/port.

    If a sentinel master name is provided then it is assumed the connection is
    to a Redis sentinel and the master and replica clients will be obtained
    from the sentinel.

    Args:
        host (str | None): The hostname of the Redis server to use.
        port (int | None): The port that Redis is served on.
        password (str | None): The Redis database password to use.
        db (int): The Redis database number to use.
        dns-srv-record (str): Domain name of SRV record used to discover host/port
        sentinel-master-name (str): Service name of Redis master instance, used
            in a Redis sentinel configuration
        retries (int): Max number of times to retry (default 3). Backoff between retries is about 2^(N-1),
            where N is the number of attempts

    Raises:
        RuntimeError: when unable to resolve a host/port to connect to

    """

    def __init__(
        self,
        *,
        host: Optional[str] = None,
        port: Optional[int] = None,
        password: Optional[str] = None,
        db: Optional[int] = None,
        dns_srv_record: Optional[str] = None,
        sentinel_master_name: Optional[str] = None,
        retries: int = 3,
    ) -> None:
        self._password = password
        self._db = db
        self._dns_srv_record = dns_srv_record
        self._sentinel_master_name = sentinel_master_name
        self._retries = retries
        self._retriable_errors = (RedisConnectionError, RedisTimeoutError)

        self._host_infos: List[Tuple[str, int]] = []
        self._sentinel: Optional[Sentinel] = None

        self._socket_timeout = 1.0

        if host is not None and port is not None:
            self._host_infos = [(host, port)]

        if self._dns_srv_record is not None:
            self._host_infos = self._perform_dns_srv_request(self._dns_srv_record)

        if not self._host_infos:
            raise RuntimeError("Either host/port or dns_srv_record must be specified")

        self._conns = self._connect()

    def _perform_dns_srv_request(self, domain_name: str) -> List[Tuple[str, int]]:
        srv_list: List[Tuple[str, int]] = []

        try:
            srv_records = dns.resolver.resolve(domain_name, "SRV")
        except Exception:
            LOGGER.debug("Unable to resolve DNS name.")
            raise RuntimeError

        for srv in srv_records:
            srv_list.append((str(srv.target).rstrip("."), srv.port))

        if not srv_list:
            raise RuntimeError("Host/port not resolvable from DNS SRV record")

        return srv_list

    def _connect(self) -> Tuple["redis.Redis[bytes]", "redis.Redis[bytes]"]:
        if self._sentinel_master_name is None:
            r = redis.Redis(
                host=self._host_infos[0][0],
                port=self._host_infos[0][1],
                socket_timeout=self._socket_timeout,
                db=self._db,  # type: ignore
                password=self._password,
            )
            return (r, r)
        else:
            if not self._sentinel:
                self._sentinel = Sentinel(
                    self._host_infos,
                    socket_timeout=self._socket_timeout,
                    db=self._db,
                    password=self._password,
                )
            return (
                self._sentinel.master_for(self._sentinel_master_name, socket_timeout=self._socket_timeout),
                self._sentinel.slave_for(self._sentinel_master_name, socket_timeout=self._socket_timeout),
            )

    def _reresolve_reconnect(self) -> None:
        if self._dns_srv_record:
            self._host_infos = self._perform_dns_srv_request(self._dns_srv_record)
        self._conns = self._connect()

    def execute_rw(self, func: Callable[["redis.Redis[bytes]"], T]) -> T:
        """Calls ``func`` with the redis read/write client as argument.

        The ``func`` may be called more than once if the host has changed.
        """
        retry_num = 0
        backoff = EqualJitterBackoff()
        while True:
            try:
                return func(self._conns[0])
            except self._retriable_errors:
                retry_num += 1
                if retry_num > self._retries:
                    raise
                sleep(backoff.compute(retry_num))
                self._reresolve_reconnect()

    def execute_ro(self, func: Callable[["redis.Redis[bytes]"], T]) -> T:
        """Calls ``func`` with a redis read-only client as argument.

        The ``func`` may be called more than once if the host has changed.
        """
        retry_num = 0
        backoff = EqualJitterBackoff()
        while True:
            try:
                return func(self._conns[1])
            except self._retriable_errors:
                retry_num += 1
                if retry_num > self._retries:
                    raise
                sleep(backoff.compute(retry_num))
                self._reresolve_reconnect()



# Copyright (C) 2022 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asyncio import Lock
from collections import OrderedDict
from datetime import datetime, timedelta
from typing import Generic, Optional, TypeVar

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import ActionResult
from buildgrid._protos.google.longrunning.operations_pb2 import Operation

T = TypeVar("T", ActionResult, Operation, bytes)


class _CacheEntry(Generic[T]):
    def __init__(self, value: T, ttl: int) -> None:
        self._never_expire = ttl <= 0
        self._expiry_date = datetime.utcnow() + timedelta(seconds=ttl)
        self.value: T = value

    def __str__(self) -> str:
        return str(self.value)

    def __repr__(self) -> str:
        return repr(self.value)

    def is_fresh(self) -> bool:
        return self._never_expire or datetime.utcnow() < self._expiry_date


class LruCache(Generic[T]):
    """Class implementing an async-safe LRU cache with an asynchronous API.

    This class provides LRU functionality similar to the existing LruCache
    class, but providing an asynchronous API and using asynchronous locks
    internally. This allows it to be used in asyncio coroutines without
    blocking the event loop.

    This class also supports setting a TTL on cache entries. Cleanup of
    entries which have outlived their TTL is done lazily when ``LruCache.get``
    is called for the relevant key.

    """

    def __init__(self, max_length: int):
        """Initialize a new LruCache.

        Args:
            max_length (int): The maximum number of entries to store in the
                cache at any time.

        """
        self._cache: "OrderedDict[str, _CacheEntry[T]]" = OrderedDict()
        self._lock = Lock()
        self._max_length = max_length

    def max_size(self) -> int:
        """Get the maximum number of items that can be stored in the cache.

        Calling ``LruCache.set`` when there are already this many entries
        in the cache will cause the oldest entry to be dropped.

        Returns:
            int: The maximum number of items to store.

        """
        return self._max_length

    async def size(self) -> int:
        """Get the current number of items in the cache.

        Returns:
            int: The number of items currently stored.

        """
        async with self._lock:
            return len(self._cache)

    async def get(self, key: str) -> Optional[T]:
        """Get the value for a given key, or ``None``.

        This method returns the value for a given key. If the key isn't
        in the cache, or the value's TTL has expired, then this returns
        ``None`` instead. In the case of a TTL expiry, the key is removed
        from the cache to save space.

        Args:
            key (str): The key to get the corresponding value for.

        Returns:
            A value mapped to the given key, or ``None``

        """
        async with self._lock:
            entry = self._cache.get(key)
            if entry is not None:
                if entry.is_fresh():
                    self._cache.move_to_end(key)
                    return entry.value
                else:
                    del self._cache[key]
        return None

    async def set(self, key: str, value: T, ttl: int) -> None:
        """Set the value and TTL of a key.

        This method sets the value and TTL of a key. A TTL of 0
        or lower means that the key won't expire due to age. Keys
        with a TTL of 0 or lower will still be dropped when they're
        the least-recently-used key.

        Args:
            key (str): The key to update.
            value (T): The value to map to the key.
            ttl (int): A TTL (in seconds) for the key.

        """
        async with self._lock:
            while len(self._cache) >= self._max_length:
                self._cache.popitem(last=False)
            self._cache[key] = _CacheEntry(value, ttl)



# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
IntrospectionService
====================

Definition of a IntrospectionService used for introspection and querying
of BuildGrid's internal state.

"""


from typing import cast

import grpc

from buildgrid._protos.buildgrid.v2.introspection_pb2 import (
    DESCRIPTOR,
    GetOperationFiltersRequest,
    ListWorkersRequest,
    ListWorkersResponse,
    OperationFilters,
)
from buildgrid._protos.buildgrid.v2.introspection_pb2_grpc import (
    IntrospectionServicer,
    add_IntrospectionServicer_to_server,
)
from buildgrid.server.decorators import rpc
from buildgrid.server.introspection.instance import IntrospectionInstance
from buildgrid.server.servicer import InstancedServicer


class IntrospectionService(IntrospectionServicer, InstancedServicer[IntrospectionInstance]):
    SERVICE_NAME = "Introspection"
    REGISTER_METHOD = add_IntrospectionServicer_to_server
    FULL_NAME = DESCRIPTOR.services_by_name["Introspection"].full_name

    @rpc(instance_getter=lambda r: cast(str, r.instance_name))
    def ListWorkers(self, request: ListWorkersRequest, context: grpc.ServicerContext) -> ListWorkersResponse:
        return self.current_instance.list_workers(request)

    @rpc(instance_getter=lambda r: cast(str, r.instance_name))
    def GetOperationFilters(
        self, request: GetOperationFiltersRequest, context: grpc.ServicerContext
    ) -> OperationFilters:
        return self.current_instance.get_operation_filters()



# Copyright (C) 2019, 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Write Once Action Cache
=======================

An ActionCache backend implementation which only allows each digest to
be written to once. Any subsequent requests to cache a result for the
same digests will not be permitted.

This can be used to wrap a different ActionCache implementation to provide
immutability of cache results.

"""

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import ActionResult, Digest
from buildgrid.server.actioncache.caches.action_cache_abc import ActionCacheABC
from buildgrid.server.exceptions import NotFoundError, UpdateNotAllowedError
from buildgrid.server.logging import buildgrid_logger

LOGGER = buildgrid_logger(__name__)


class WriteOnceActionCache(ActionCacheABC):
    def __init__(self, action_cache: ActionCacheABC):
        super().__init__()
        self._action_cache = action_cache

    def start(self) -> None:
        self._action_cache.start()

    def stop(self) -> None:
        self._action_cache.stop()

    @property
    def allow_updates(self) -> bool:
        return self._action_cache.allow_updates

    def get_action_result(self, action_digest: Digest) -> ActionResult:
        """Retrieves the cached ActionResult for the given Action digest.

        Args:
            action_digest (Digest): The digest of the Action to retrieve the
                cached result of.

        Returns:
            The cached ActionResult matching the given digest or raises
            NotFoundError.

        """
        return self._action_cache.get_action_result(action_digest)

    def update_action_result(self, action_digest: Digest, action_result: ActionResult) -> None:
        """Stores the result for a given digest in the cache.

        If the digest already exists in the cache, then an UpdateNotAllowedError
        is raised instead of storing the result.

        Args:
            action_digest (Digest): The digest of the Action whose result is
                being cached.
            action_result (ActionResult): The result to cache for the given
                Action digest.

        """
        try:
            self._action_cache.get_action_result(action_digest)
            # This should throw NotFoundError or actually exist
            LOGGER.warning(
                "Result already cached for action, WriteOnceActionCache won't overwrite it to the new action result.",
                tags=dict(digest=action_digest, action_result=action_result),
            )

            raise UpdateNotAllowedError(
                "Result already stored for this action digest;WriteOnceActionCache doesn't allow updates."
            )
        except NotFoundError:
            self._action_cache.update_action_result(action_digest, action_result)



# Copyright (C) 2018 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

from buildgrid._protos.build.bazel.remote.execution.v2 import remote_execution_pb2
from buildgrid.server.decorators.requestid import track_request_id
from buildgrid.server.metadata import ctx_grpc_request_id
from buildgrid.server.utils.digests import create_digest, get_hash_type, parse_digest

BLOBS = (
    b"",
    b"non-empty-blob",
)
BLOB_HASHES = (
    "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
    "89070dfb3175a2c75835d70147b52bd97afd8228819566d84eecd2d20e9b19fc",
)
BLOB_SIZES = (
    0,
    14,
)
BLOB_DATA = zip(BLOBS, BLOB_HASHES, BLOB_SIZES)

STRINGS = (
    "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855/0",
    "89070dfb3175a2c75835d70147b52bd97afd8228819566d84eecd2d20e9b19fc/14",
    "e1ca41574914ba00e8ed5c8fc78ec8efdfd48941c7e48ad74dad8ada7f2066d/12",
)
BLOB_HASHES = (
    "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
    "89070dfb3175a2c75835d70147b52bd97afd8228819566d84eecd2d20e9b19fc",
    None,
)
BLOB_SIZES = (
    0,
    14,
    None,
)
STRING_VALIDITIES = (
    True,
    True,
    False,
)
STRING_DATA = zip(STRINGS, BLOB_HASHES, BLOB_SIZES, STRING_VALIDITIES)

BASE_URL = "http://localhost:8080"
INSTANCES = (
    None,
    "",
    "instance",
)
URL_HASHES = (
    "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
    "89070dfb3175a2c75835d70147b52bd97afd8228819566d84eecd2d20e9b19fc",
)
URL_SIZES = (
    0,
    14,
)
URL_DATA = zip(URL_HASHES, URL_SIZES)


@pytest.mark.parametrize("blob,digest_hash,digest_size", BLOB_DATA)
def test_create_digest(blob, digest_hash, digest_size):
    # Generate a Digest message from given blob:
    blob_digest = create_digest(blob)

    assert get_hash_type() == remote_execution_pb2.DigestFunction.SHA256

    assert hasattr(blob_digest, "DESCRIPTOR")
    assert blob_digest.DESCRIPTOR == remote_execution_pb2.Digest.DESCRIPTOR
    assert blob_digest.hash == digest_hash
    assert blob_digest.size_bytes == digest_size


@pytest.mark.parametrize("string,digest_hash,digest_size,validity", STRING_DATA)
def test_parse_digest(string, digest_hash, digest_size, validity):
    # Generate a Digest message from given string:
    string_digest = parse_digest(string)

    assert get_hash_type() == remote_execution_pb2.DigestFunction.SHA256

    if validity:
        assert hasattr(string_digest, "DESCRIPTOR")
        assert string_digest.DESCRIPTOR == remote_execution_pb2.Digest.DESCRIPTOR
        assert string_digest.hash == digest_hash
        assert string_digest.size_bytes == digest_size

    else:
        assert string_digest is None


def test_track_request_id():
    @track_request_id
    def _example_request_handler():
        assert ctx_grpc_request_id.get() is not None

    assert ctx_grpc_request_id.get() is None
    _example_request_handler()
    assert ctx_grpc_request_id.get() is None


def test_request_id_unset_on_error():
    @track_request_id
    def _example_request_handler():
        assert ctx_grpc_request_id.get() is not None
        raise RuntimeError("expected")

    assert ctx_grpc_request_id.get() is None
    with pytest.raises(RuntimeError):
        _example_request_handler()
    assert ctx_grpc_request_id.get() is None


def test_track_request_id_generator():
    @track_request_id
    def _example_request_handler():
        assert ctx_grpc_request_id.get() is not None
        yield

    assert ctx_grpc_request_id.get() is None
    for _ in _example_request_handler():
        pass
    assert ctx_grpc_request_id.get() is None


def test_request_id_generator_unset_on_error():
    @track_request_id
    def _example_generator_handler():
        assert ctx_grpc_request_id.get() is not None
        yield
        raise RuntimeError("expected")

    assert ctx_grpc_request_id.get() is None
    with pytest.raises(RuntimeError):
        for _ in _example_generator_handler():
            assert ctx_grpc_request_id.get() is not None
    assert ctx_grpc_request_id.get() is None



# Copyright (C) 2020 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""digest_size_bytes to bigint

Revision ID: c64c104b2c8b
Revises: 29e88b3e0d0a
Create Date: 2021-10-18 15:02:56.155398

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "c64c104b2c8b"
down_revision = "29e88b3e0d0a"
branch_labels = None
depends_on = None


def upgrade() -> None:
    with op.batch_alter_table("index") as batch_op:
        batch_op.alter_column("digest_size_bytes", type_=sa.BigInteger())


def downgrade() -> None:
    with op.batch_alter_table("index") as batch_op:
        batch_op.alter_column("digest_size_bytes", type_=sa.Integer())



# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
ReplicatedStorage
=========================

A storage provider which stores data in multiple storages, replicating
any data missing in some but present in others.

"""


from contextlib import ExitStack
from typing import IO, Dict, List, Optional, Set, Tuple

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import Digest
from buildgrid._protos.google.rpc import code_pb2
from buildgrid._protos.google.rpc.status_pb2 import Status
from buildgrid.server.decorators import timed
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.metrics_names import METRIC
from buildgrid.server.metrics_utils import publish_counter_metric
from buildgrid.server.utils.digests import HashableDigest

from .storage_abc import StorageABC

LOGGER = buildgrid_logger(__name__)


class ReplicatedStorage(StorageABC):
    TYPE = "Replicated"

    def __init__(self, storages: List[StorageABC]) -> None:
        self._stack = ExitStack()
        self._storages = dict(enumerate(storages))

    def start(self) -> None:
        for storage in self._storages.values():
            self._stack.enter_context(storage)

    def stop(self) -> None:
        self._stack.close()

    @timed(METRIC.STORAGE.STAT_DURATION, type=TYPE)
    def has_blob(self, digest: Digest) -> bool:
        has_blob: Set[int] = set(i for i in self._storages if self._storages[i].has_blob(digest))
        missing_blob = set(self._storages.keys()) - has_blob
        if len(missing_blob) < len(self._storages):
            publish_counter_metric(METRIC.STORAGE.REPLICATED.REQUIRED_REPLICATION_COUNT, len(missing_blob))
        return len(has_blob) > 0

    @timed(METRIC.STORAGE.READ_DURATION, type=TYPE)
    def get_blob(self, digest: Digest) -> Optional[IO[bytes]]:
        has_blob: Set[int] = set(i for i in self._storages if self._storages[i].has_blob(digest))
        missing_blob = set(self._storages.keys()) - has_blob
        blob = None
        failed_writes = 0
        for idx in has_blob:
            if blob := self._storages[idx].get_blob(digest):
                break
            LOGGER.error(
                "Storage shard reported digest exists but downloading failed.",
                tags=dict(shard_index=idx, digest=digest),
            )
            missing_blob.add(idx)
        if len(missing_blob) < len(self._storages):
            assert blob is not None
            for idx in missing_blob:
                try:
                    self._storages[idx].commit_write(digest, blob)
                    LOGGER.debug("Replicated digest to storage shard.", tags=dict(shard_index=idx, digest=digest))
                except Exception as e:
                    LOGGER.warning(
                        f"Failed to replicate digest to storage shard: {e}.", tags=dict(shard_index=idx, digest=digest)
                    )
                    failed_writes += 1
                blob.seek(0)

            publish_counter_metric(METRIC.STORAGE.REPLICATED.REPLICATION_ERROR_COUNT, failed_writes)
            publish_counter_metric(METRIC.STORAGE.REPLICATED.REPLICATION_COUNT, len(missing_blob) - failed_writes)
        return blob

    @timed(METRIC.STORAGE.DELETE_DURATION, type=TYPE)
    def delete_blob(self, digest: Digest) -> None:
        for storage in self._storages.values():
            storage.delete_blob(digest)

    @timed(METRIC.STORAGE.WRITE_DURATION, type=TYPE)
    def commit_write(self, digest: Digest, write_session: IO[bytes]) -> None:
        failed_writes = 0
        error_msgs: List[str] = []
        for idx, storage in self._storages.items():
            try:
                storage.commit_write(digest, write_session)
            except Exception as error:
                LOGGER.warning(
                    f"Failed to write digest to storage shard: {error}", tags=dict(shard_index=idx, digest=digest)
                )
                error_msgs.append(str(error))
                failed_writes += 1
            write_session.seek(0)

        publish_counter_metric(METRIC.STORAGE.REPLICATED.REPLICATION_ERROR_COUNT, failed_writes)
        if failed_writes == len(self._storages):
            error_string = "Writes to all storages failed with the following errors:\n"
            error_string += "\n".join(error_msgs)
            LOGGER.error(error_string)
            raise RuntimeError(error_string)

    @timed(METRIC.STORAGE.BULK_DELETE_DURATION, type=TYPE)
    def bulk_delete(self, digests: List[Digest]) -> List[str]:
        failed_deletions_set: Set[str] = set()
        for storage in self._storages.values():
            failed_deletions_set.union(storage.bulk_delete(digests))
        publish_counter_metric(METRIC.STORAGE.DELETE_ERRORS_COUNT, len(failed_deletions_set), type=self.TYPE)
        return list(failed_deletions_set)

    @timed(METRIC.STORAGE.BULK_STAT_DURATION, type=TYPE)
    def missing_blobs(self, digests: List[Digest]) -> List[Digest]:
        """Call missing_blobs on each storage and only report a blob is missing if it's in none of the
        storages. The number of blobs missing from a storage but present in others is logged and published
        as a metric
        """
        digest_set: Set[HashableDigest] = {HashableDigest(digest.hash, digest.size_bytes) for digest in digests}
        missing_from_all_storages = digest_set.copy()
        missing_for_storage: Dict[int, Set[HashableDigest]] = {}

        blobs_needing_replication_count = 0
        for idx, storage in self._storages.items():
            response = storage.missing_blobs(digests)
            missing_for_storage[idx] = {HashableDigest(digest.hash, digest.size_bytes) for digest in response}

        missing_from_all_storages = set.intersection(*missing_for_storage.values())
        digests_in_at_least_one_storage: Set[HashableDigest] = digest_set - missing_from_all_storages
        # Figure out what blobs are present in some storages but not all to report metrics
        for idx, missing_digests in missing_for_storage.items():
            inconsistent_digests_for_storage = missing_digests & digests_in_at_least_one_storage
            if inconsistent_digests_for_storage:
                LOGGER.info(
                    "Storage shard has blobs which need to be replicated.",
                    tags=dict(shard_index=idx, digest_count=len(inconsistent_digests_for_storage)),
                )
                blobs_needing_replication_count += len(inconsistent_digests_for_storage)

        publish_counter_metric(METRIC.STORAGE.REPLICATED.REQUIRED_REPLICATION_COUNT, blobs_needing_replication_count)
        missing_blobs_list = [hdigest.to_digest() for hdigest in missing_from_all_storages]
        return missing_blobs_list

    # Bulk write to all storages. Errors are not fatal as long as one storage is
    # successfully written to
    @timed(METRIC.STORAGE.BULK_WRITE_DURATION, type=TYPE)
    def bulk_update_blobs(self, blobs: List[Tuple[Digest, bytes]]) -> List[Status]:
        digest_result: Dict[HashableDigest, Status] = {}
        errored_blobs_total = 0
        for idx, storage in self._storages.items():
            errored_blobs_for_storage = 0
            results = storage.bulk_update_blobs(blobs)
            for digest_blob_tuple, result in zip(blobs, results):
                digest, _ = digest_blob_tuple
                hdigest = HashableDigest(hash=digest.hash, size_bytes=digest.size_bytes)

                if result.code != code_pb2.OK:
                    errored_blobs_for_storage += 1

                # Keep track of the status code for this digest, preferring OK over errors
                if hdigest not in digest_result or digest_result[hdigest].code != code_pb2.OK:
                    digest_result[hdigest] = result

            if errored_blobs_for_storage > 0:
                LOGGER.warning(
                    "Failed to write all digests to storage shard.",
                    tags=dict(shard_index=idx, digest_count=len(results), error_count=errored_blobs_for_storage),
                )
                errored_blobs_total += errored_blobs_for_storage

        publish_counter_metric(METRIC.STORAGE.REPLICATED.REPLICATION_ERROR_COUNT, errored_blobs_total)
        return [digest_result[hdigest] for hdigest in digest_result]

    # Read blobs from all storages, writing any missing blobs into storages missing
    # them from storages that have them
    @timed(METRIC.STORAGE.BULK_READ_DURATION, type=TYPE)
    def bulk_read_blobs(self, digests: List[Digest]) -> Dict[str, bytes]:
        digest_set = set(HashableDigest(hash=digest.hash, size_bytes=digest.size_bytes) for digest in digests)
        missing_blobs: Dict[int, Set[HashableDigest]] = {}
        bulk_read_results: Dict[str, bytes] = {}
        # Find what blobs are missing for this storage and read what's available
        for idx, storage in self._storages.items():
            missing_blobs[idx] = set(
                [
                    HashableDigest(hash=digest.hash, size_bytes=digest.size_bytes)
                    for digest in storage.missing_blobs(digests)
                ]
            )
            present_blobs = digest_set - missing_blobs[idx]
            blobs_to_read = [x.to_digest() for x in present_blobs if x.hash not in bulk_read_results]
            bulk_read_results.update(self._storages[idx].bulk_read_blobs(blobs_to_read))

        replicated_blobs_count = 0
        errored_blobs_count = 0
        for idx, missing in missing_blobs.items():
            # Write any blobs that exist in other storages which are missing from this storage
            write_batch: List[Tuple[Digest, bytes]] = []
            for digest in missing:
                if digest.hash in bulk_read_results:
                    write_batch.append((digest.to_digest(), bulk_read_results[digest.hash]))
            if write_batch:
                update_results = self._storages[idx].bulk_update_blobs(write_batch)
                for result in update_results:
                    if result.code != code_pb2.OK:
                        errored_blobs_count += 1
                    else:
                        replicated_blobs_count += 1
                LOGGER.debug(
                    "Replicated blobs to storage shard.", tags=dict(shard_index=idx, digest_count=len(write_batch))
                )

        publish_counter_metric(METRIC.STORAGE.REPLICATED.REPLICATION_COUNT, replicated_blobs_count)
        publish_counter_metric(METRIC.STORAGE.REPLICATED.REPLICATION_ERROR_COUNT, errored_blobs_count)
        return bulk_read_results



# Copyright (C) 2021 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License' is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import json
import os
import shutil
import tarfile
import tempfile
from collections import namedtuple
from typing import Any, Awaitable, Callable, Dict, Optional, Tuple, Type

import aiofiles
from aiohttp import WSMsgType, web
from aiohttp_middlewares.annotations import UrlCollection
from grpc import RpcError, StatusCode
from grpc.aio import Call, Metadata

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import (
    ActionResult,
    Digest,
    Directory,
    GetActionResultRequest,
    RequestMetadata,
)
from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2_grpc_aio import ActionCacheStub
from buildgrid._protos.buildgrid.v2.query_build_events_pb2 import QueryEventStreamsRequest
from buildgrid._protos.buildgrid.v2.query_build_events_pb2_grpc_aio import QueryBuildEventsStub
from buildgrid._protos.google.bytestream.bytestream_pb2 import ReadRequest
from buildgrid._protos.google.bytestream.bytestream_pb2_grpc_aio import ByteStreamStub
from buildgrid._protos.google.longrunning import operations_pb2
from buildgrid._protos.google.longrunning.operations_pb2_grpc_aio import OperationsStub
from buildgrid.server.app.cli import Context
from buildgrid.server.browser.utils import TARBALL_DIRECTORY_PREFIX, ResponseCache, get_cors_headers
from buildgrid.server.logging import buildgrid_logger
from buildgrid.server.metadata import extract_request_metadata, extract_trailing_client_identity
from buildgrid.server.operations.filtering.interpreter import VALID_OPERATION_FILTERS, OperationFilterSpec
from buildgrid.server.operations.filtering.sanitizer import DatetimeValueSanitizer, SortKeyValueSanitizer
from buildgrid.server.settings import BROWSER_MAX_CACHE_ENTRY_SIZE

LOGGER = buildgrid_logger(__name__)


def query_build_events_handler(context: Context) -> Callable[[web.Request], Awaitable[web.Response]]:
    """Factory function which returns a handler for QueryEventStreams.

    The returned handler uses ``context.channel`` to send a QueryEventStreams
    request constructed based on the provided URL query parameters. Currently
    only querying by build_id (equivalent to correlated invocations ID) is
    supported.

    The handler returns a serialised QueryEventStreamsResponse, and raises a
    500 error in the case of some RPC error.

    Args:
        context (Context): The context to use to send the gRPC request.

    """

    async def _query_build_events_handler(request: web.Request) -> web.Response:
        build_id = request.rel_url.query.get("build_id", ".*")

        stub = QueryBuildEventsStub(context.channel)  # type: ignore # Requires stub regen

        grpc_request = QueryEventStreamsRequest(build_id_pattern=build_id)

        try:
            grpc_response = await stub.QueryEventStreams(grpc_request)
        except RpcError as e:
            LOGGER.warning(e.details())
            raise web.HTTPInternalServerError()

        serialized_response = grpc_response.SerializeToString()
        return web.Response(body=serialized_response)

    return _query_build_events_handler


async def get_operation_filters_handler(request: web.Request) -> web.Response:
    """Return the available Operation filter keys."""

    def _generate_filter_spec(key: str, spec: OperationFilterSpec) -> Dict[str, Any]:
        comparators = ["<", "<=", "=", "!=", ">=", ">"]
        filter_type = "text"
        if isinstance(spec.sanitizer, SortKeyValueSanitizer):
            comparators = ["="]
        elif isinstance(spec.sanitizer, DatetimeValueSanitizer):
            filter_type = "datetime"

        ret = {
            "comparators": comparators,
            "description": spec.description,
            "key": key,
            "name": spec.name,
            "type": filter_type,
        }

        try:
            ret["values"] = spec.sanitizer.valid_values
        except NotImplementedError:
            pass
        return ret

    operation_filters = [_generate_filter_spec(key, spec) for key, spec in VALID_OPERATION_FILTERS.items()]
    return web.Response(text=json.dumps(operation_filters))


def list_operations_handler(
    context: Context, cache: ResponseCache
) -> Callable[[web.Request], Awaitable[web.Response]]:
    """Factory function which returns a handler for ListOperations.

    The returned handler uses ``context.channel`` and ``context.instance_name``
    to send a ListOperations request constructed based on the provided URL
    query parameters.

    The handler returns a serialised ListOperationsResponse, raises a 400
    error in the case of a bad filter or other invalid argument, or raises
    a 500 error in the case of some other RPC error.

    Args:
        context (Context): The context to use to send the gRPC request.

    """

    async def _list_operations_handler(request: web.Request) -> web.Response:
        filter_string = request.rel_url.query.get("q", "")
        page_token = request.rel_url.query.get("page_token", "")
        page_size_str = request.rel_url.query.get("page_size")
        page_size = 0
        if page_size_str is not None:
            page_size = int(page_size_str)

        LOGGER.info(
            "Received ListOperations request.",
            tags=dict(filter_string=filter_string, page_token=page_token, page_size=page_size),
        )
        stub = OperationsStub(context.operations_channel)  # type: ignore # Requires stub regen
        grpc_request = operations_pb2.ListOperationsRequest(
            name=context.instance_name, page_token=page_token, page_size=page_size, filter=filter_string
        )

        try:
            grpc_response = await stub.ListOperations(grpc_request)
        except RpcError as e:
            LOGGER.warning(e.details())
            if e.code() == StatusCode.INVALID_ARGUMENT:
                raise web.HTTPBadRequest()
            raise web.HTTPInternalServerError()

        serialised_response = grpc_response.SerializeToString()
        return web.Response(body=serialised_response)

    return _list_operations_handler


async def _get_operation(context: Context, request: web.Request) -> Tuple[operations_pb2.Operation, Call]:
    operation_name = f"{context.instance_name}/{request.match_info['name']}"

    stub = OperationsStub(context.operations_channel)  # type: ignore # Requires stub regen
    grpc_request = operations_pb2.GetOperationRequest(name=operation_name)

    try:
        call = stub.GetOperation(grpc_request)
        operation = await call
    except RpcError as e:
        LOGGER.warning(f"Error fetching operation: {e.details()}")
        if e.code() == StatusCode.INVALID_ARGUMENT:
            raise web.HTTPNotFound()
        raise web.HTTPInternalServerError()

    return operation, call


def get_operation_handler(context: Context, cache: ResponseCache) -> Callable[[web.Request], Awaitable[web.Response]]:
    """Factory function which returns a handler for GetOperation.

    The returned handler uses ``context.channel`` and ``context.instance_name``
    to send a GetOperation request constructed based on the path component of
    the URL.

    The handler returns a serialised Operation message, raises a 400 error in
    the case of an invalid operation name, or raises a 500 error in the case
    of some other RPC error.

    Args:
        context (Context): The context to use to send the gRPC request.

    """

    async def _get_operation_handler(request: web.Request) -> web.Response:
        name = request.match_info["name"]
        LOGGER.info("Received GetOperation request.", tags=dict(name=name))
        operation, _ = await _get_operation(context, request)

        serialised_response = operation.SerializeToString()
        return web.Response(body=serialised_response)

    return _get_operation_handler


def get_operation_request_metadata_handler(context: Context) -> Callable[[web.Request], Awaitable[web.Response]]:
    """Factory function which returns a handler to get RequestMetadata.

    The returned handler uses ``context.channel`` and ``context.instance_name``
    to send a GetOperation request constructed based on the path component of
    the URL.

    The handler returns a serialised RequestMetadata proto message, retrieved
    from the trailing metadata of the GetOperation response. In the event of
    an invalid operation name it raises a 404 error, and raises a 500 error in
    the case of some other RPC error.

    Args:
        context (Context): The context to use to send the gRPC request.

    """

    async def _get_operation_request_metadata_handler(request: web.Request) -> web.Response:
        LOGGER.info("Received request for RequestMetadata.", tags=dict(name=str(request.match_info["name"])))
        _, call = await _get_operation(context, request)
        metadata = await call.trailing_metadata()

        def extract_metadata(m: Metadata) -> RequestMetadata:
            # `m` contains a list of tuples, but `extract_request_metadata()`
            # expects a `key` and `value` attributes.
            MetadataTuple = namedtuple("MetadataTuple", ["key", "value"])
            return extract_request_metadata([MetadataTuple(entry[0], entry[1]) for entry in m])

        request_metadata = extract_metadata(metadata)
        return web.Response(body=request_metadata.SerializeToString())

    return _get_operation_request_metadata_handler


def get_operation_client_identity_handler(context: Context) -> Callable[[web.Request], Awaitable[web.Response]]:
    """Factory function which returns a handler to get ClientIdentity metadata.

    The returned handler uses ``context.channel`` and ``context.instance_name``
    to send a GetOperation request constructed based on the path component of
    the URL.

    The handler returns a serialised ClientIdentity proto message, retrieved
    from the trailing metadata of the GetOperation response. In the event of
    an invalid operation name it raises a 404 error, and raises a 500 error in
    the case of some other RPC error.

    Args:
        context (Context): The context to use to send the gRPC request.

    """

    async def _get_operation_client_identity_handler(request: web.Request) -> web.Response:
        LOGGER.info("Received request for RequestMetadata.", tags=dict(name=str(request.match_info["name"])))
        _, call = await _get_operation(context, request)
        metadata = await call.trailing_metadata()
        client_identity = extract_trailing_client_identity(metadata)
        return web.Response(body=client_identity.SerializeToString())

    return _get_operation_client_identity_handler


def cancel_operation_handler(context: Context) -> Callable[[web.Request], Awaitable[web.Response]]:
    """Factory function which returns a handler for CancelOperation.

    The returned handler uses ``context.channel`` and ``context.instance_name``
    to send a CancelOperation request constructed based on the path component of
    the URL.

    The handler raises a 404 error in the case of an invalid operation name,
    or a 500 error in the case of some other RPC error.

    On success, the response is empty.

    Args:
        context (Context): The context to use to send the gRPC request.

    """

    async def _cancel_operation_handler(request: web.Request) -> web.Response:
        LOGGER.info("Received CancelOperation request.", tags=dict(name=str(request.match_info["name"])))
        operation_name = f"{context.instance_name}/{request.match_info['name']}"

        stub = OperationsStub(context.operations_channel)  # type: ignore # Requires stub regen
        grpc_request = operations_pb2.CancelOperationRequest(name=operation_name)

        try:
            await stub.CancelOperation(grpc_request)
            return web.Response()
        except RpcError as e:
            LOGGER.warning(e.details())
            if e.code() == StatusCode.INVALID_ARGUMENT:
                raise web.HTTPNotFound()
            raise web.HTTPInternalServerError()

    return _cancel_operation_handler


async def _fetch_action_result(
    context: Context, request: web.Request, cache: ResponseCache, cache_key: str
) -> ActionResult:
    stub = ActionCacheStub(context.cache_channel)  # type: ignore # Requires stub regen
    digest = Digest(hash=request.match_info["hash"], size_bytes=int(request.match_info["size_bytes"]))
    grpc_request = GetActionResultRequest(action_digest=digest, instance_name=context.instance_name)

    try:
        result = await stub.GetActionResult(grpc_request)
    except RpcError as e:
        LOGGER.warning(f"Failed to fetch ActionResult: [{e.details()}]")
        if e.code() == StatusCode.NOT_FOUND:
            raise web.HTTPNotFound()
        raise web.HTTPInternalServerError()

    await cache.store_action_result(cache_key, result)
    return result


def get_action_result_handler(
    context: Context, cache: ResponseCache, cache_capacity: int = 512
) -> Callable[[web.Request], Awaitable[web.Response]]:
    """Factory function which returns a handler for GetActionResult.

    The returned handler uses ``context.channel`` and ``context.instance_name``
    to send a GetActionResult request constructed based on the path components
    of the URL.

    The handler returns a serialised ActionResult message, raises a 404 error
    if there's no result cached, or raises a 500 error in the case of some
    other RPC error.

    Args:
        context (Context): The context to use to send the gRPC request.
        cache_capacity (int): The number of ActionResults to cache in memory
            to avoid hitting the actual ActionCache.

    """

    class FetchSpec:
        """Simple class used to store information about a GetActionResult request.

        A class is used here rather than a namedtuple since we need this state
        to be mutable.

        """

        def __init__(
            self, *, error: Optional[Exception], event: asyncio.Event, result: Optional[ActionResult], refcount: int
        ):
            self.error = error
            self.event = event
            self.result = result
            self.refcount = refcount

    in_flight_fetches: Dict[str, FetchSpec] = {}
    fetch_lock = asyncio.Lock()

    async def _get_action_result_handler(request: web.Request) -> web.Response:
        cache_key = f'{request.match_info["hash"]}/{request.match_info["size_bytes"]}'
        LOGGER.info("Received GetActionResult request.", tags=dict(cache_key=cache_key))

        result = await cache.get_action_result(cache_key)

        if result is None:
            try:
                duplicate_request = False
                spec = None
                async with fetch_lock:
                    if cache_key in in_flight_fetches:
                        LOGGER.info("Deduplicating GetActionResult request.", tags=dict(cache_key=cache_key))
                        spec = in_flight_fetches[cache_key]
                        spec.refcount += 1
                        duplicate_request = True
                    else:
                        spec = FetchSpec(error=None, event=asyncio.Event(), result=None, refcount=1)
                        in_flight_fetches[cache_key] = spec

                if duplicate_request and spec:
                    # If we're a duplicate of an existing request, then wait for the
                    # existing request to finish fetching from the ActionCache.
                    await spec.event.wait()
                    if spec.error is not None:
                        raise spec.error
                    if spec.result is None:
                        # Note: this should be impossible, but lets guard against accidentally setting
                        # the event before the result is populated
                        LOGGER.info("Result not set in deduplicated request.", tags=dict(cache_key=cache_key))
                        raise web.HTTPInternalServerError()
                    result = spec.result
                else:
                    try:
                        result = await _fetch_action_result(context, request, cache, cache_key)
                    except Exception as e:
                        async with fetch_lock:
                            if spec is not None:
                                spec.error = e
                        raise e

                    async with fetch_lock:
                        if spec is not None:
                            spec.result = result
                            spec.event.set()

            finally:
                async with fetch_lock:
                    # Decrement refcount now we're done with the result. If we're the
                    # last request interested in the result then remove it from the
                    # `in_flight_fetches` dictionary.
                    spec = in_flight_fetches.get(cache_key)
                    if spec is not None:
                        spec.refcount -= 1
                        if spec.refcount <= 0:
                            in_flight_fetches.pop(cache_key)

        return web.Response(body=result.SerializeToString())

    return _get_action_result_handler


def get_blob_handler(
    context: Context, cache: ResponseCache, allow_all: bool = False, allowed_origins: UrlCollection = ()
) -> Callable[[web.Request], Awaitable[web.StreamResponse]]:
    async def _get_blob_handler(request: web.Request) -> web.StreamResponse:
        digest = Digest(hash=request.match_info["hash"], size_bytes=int(request.match_info["size_bytes"]))
        try:
            offset = int(request.rel_url.query.get("offset", "0"))
            limit = int(request.rel_url.query.get("limit", "0"))
        except ValueError:
            raise web.HTTPBadRequest()

        response = web.StreamResponse()

        # We need to explicitly set CORS headers here, because the middleware that
        # normally handles this only adds the header when the response is returned
        # from this function. However, when using a StreamResponse with chunked
        # encoding enabled, the client begins to receive the response when we call
        # `response.write()`. This leads to the request being disallowed due to the
        # missing reponse header for clients executing in browsers.
        cors_headers = get_cors_headers(request.headers.get("Origin"), allowed_origins, allow_all)
        response.headers.update(cors_headers)

        if request.rel_url.query.get("raw", "") == "1":
            response.headers["Content-type"] = "text/plain; charset=utf-8"
        else:
            # Setting the Content-Disposition header so that when
            # downloading a blob with a browser the default name uniquely
            # identifies its contents:
            filename = f"{request.match_info['hash']}_{request.match_info['size_bytes']}"

            # For partial reads also include the indices to prevent file mix-ups:
            if offset != 0 or limit != 0:
                filename += f"_chunk_{offset}-"
                if limit != 0:
                    filename += f"{offset + limit}"

            response.headers["Content-Disposition"] = f"Attachment;filename={filename}"

        prepared = False

        async def _callback(data: bytes) -> None:
            nonlocal prepared
            if not prepared:
                # Prepare for chunked encoding when the callback is first called,
                # so that we're sure we actually have some data before doing so.
                response.enable_chunked_encoding()
                await response.prepare(request)
                prepared = True
            await response.write(data)

        await _fetch_blob(context, cache, digest, callback=_callback, offset=offset, limit=limit)

        return response

    return _get_blob_handler


def _create_tarball(directory: str, name: str) -> bool:
    """Makes a tarball from a given directory.

    Returns True if the tarball was successfully created, and False if not.

    Args:
        directory (str): The directory to tar up.
        name (str): The name of the tarball to be produced.

    """
    try:
        with tarfile.open(name, "w:gz") as tarball:
            tarball.add(directory, arcname="")
    except Exception:
        return False
    return True


async def _fetch_blob(
    context: Context,
    cache: ResponseCache,
    digest: Digest,
    message_class: Optional[Type[Any]] = None,
    callback: Optional[Callable[[bytes], Awaitable[Any]]] = None,
    offset: int = 0,
    limit: int = 0,
) -> Any:
    """Fetch a blob from CAS.

    This function sends a ByteStream Read request for the given digest. If ``callback``
    is set then the callback is called with the data in each ReadResponse message and
    this function returns an empty bytes object once the response is finished. If
    ``message_class`` is set with no ``callback`` set then this function calls
    ``message_class.FromString`` on the fetched blob and returns the result.

    If neither ``callback`` or ``message_class`` are set then this function just returns
    the raw blob that was fetched from CAS.

    Args:
        context (Context): The context to use to send the gRPC request.
        cache (ResponseCache): The response cache to check/update with the fetched blob.
        digest (Digest): The Digest of the blob to fetch from CAS.
        message_class (type): A class which implements a ``FromString`` class method.
            The method should take a bytes object expected to contain the blob fetched
            from CAS.
        callback (callable): A function or other callable to act on a subset of the
            blob contents.
        offset (int): Read offset to start reading the blob at. Defaults to 0, the start
            of the blob.
        limit (int): Maximum number of bytes to read from the blob. Defaults to 0, no
            limit.

    """
    cacheable = digest.size_bytes <= BROWSER_MAX_CACHE_ENTRY_SIZE
    resource_name = f"{context.instance_name}/blobs/{digest.hash}/{digest.size_bytes}"
    blob = None
    if cacheable:
        blob = await cache.get_blob(resource_name)
        if blob is not None and callback is not None:
            if limit > 0:
                try:
                    blob = blob[offset : offset + limit]
                except IndexError:
                    raise web.HTTPBadRequest()
            await callback(blob)

    if blob is None:
        stub = ByteStreamStub(context.cas_channel)  # type: ignore # Requires stub regen
        grpc_request = ReadRequest(resource_name=resource_name, read_offset=offset, read_limit=limit)

        blob = b""
        try:
            async for grpc_response in stub.Read(grpc_request):
                if grpc_response.data:
                    if callback is not None:
                        await callback(grpc_response.data)

                    if callback is None or cacheable:
                        blob += grpc_response.data
        except RpcError as e:
            LOGGER.warning(e.details())
            if e.code() == StatusCode.NOT_FOUND:
                raise web.HTTPNotFound()
            raise web.HTTPInternalServerError()

        if cacheable:
            await cache.store_blob(resource_name, blob)

    if message_class is not None and callback is None:
        return message_class.FromString(blob)
    return blob


async def _download_directory(
    context: Context, cache: ResponseCache, base: str, path: str, directory: Directory
) -> None:
    """Download the contents of a directory from CAS.

    This function takes a Directory message and downloads the directory
    contents defined in the message from CAS. Raises a 400 error if the
    directory contains a symlink which points to a location outside the
    initial directory.

    The directory is downloaded recursively depth-first.

    Args:
        context (Context): The context to use for making gRPC requests.
        cache (ResponseCache): The response cache to use when downloading the
            directory contents.
        base (str): The initial directory path, used to check symlinks don't
            escape into the wider filesystem.
        path (str): The path to download the directory into.
        directory (Directory): The Directory message to fetch the contents of.

    """
    for directory_node in directory.directories:
        dir_path = os.path.join(path, directory_node.name)
        os.mkdir(dir_path)
        child = await _fetch_blob(context, cache, directory_node.digest, message_class=Directory)
        await _download_directory(context, cache, base, dir_path, child)

    for file_node in directory.files:
        file_path = os.path.join(path, file_node.name)
        async with aiofiles.open(file_path, "wb") as f:
            await _fetch_blob(context, cache, file_node.digest, callback=f.write)
        if file_node.is_executable:
            os.chmod(file_path, 0o755)

    for link_node in directory.symlinks:
        link_path = os.path.join(path, link_node.name)
        target_path = os.path.realpath(link_node.target)
        target_relpath = os.path.relpath(base, target_path)
        if target_relpath.startswith(os.pardir):
            raise web.HTTPBadRequest(
                reason="Requested directory contains a symlink targeting a location outside the tarball"
            )

        os.symlink(link_node.target, link_path)


async def _tarball_from_directory(context: Context, cache: ResponseCache, directory: Directory, tmp_dir: str) -> str:
    """Construct a tarball of a directory stored in CAS.

    This function fetches the contents of the given directory message into a
    temporary directory, and then constructs a tarball of the directory. The
    path to this tarball is returned when construction is complete.

    Args:
        context (Context): The context to use to send the gRPC requests.
        cache (ResponseCache): The response cache to use when fetching the
            tarball contents.
        directory (Directory): The Directory message for the directory we're
            making a tarball of.
        tmp_dir (str): Path to a temporary directory to use for storing the
            directory contents and its tarball.

    """
    tarball_dir = tempfile.mkdtemp(dir=tmp_dir)
    tarball_path = os.path.join(tmp_dir, "directory.tar.gz")
    loop = asyncio.get_event_loop()

    # Fetch the contents of the directory into a temporary directory
    await _download_directory(context, cache, tarball_dir, tarball_dir, directory)

    # Make a tarball from that temporary directory
    # NOTE: We do this using loop.run_in_executor to avoid the
    # synchronous and blocking tarball construction
    tarball_result = await loop.run_in_executor(None, _create_tarball, tarball_dir, tarball_path)
    if not tarball_result:
        raise web.HTTPInternalServerError()
    return tarball_path


def get_tarball_handler(
    context: Context,
    cache: ResponseCache,
    allow_all: bool = False,
    allowed_origins: UrlCollection = (),
    tarball_dir: Optional[str] = None,
) -> Callable[[web.Request], Awaitable[web.StreamResponse]]:
    """Factory function which returns a handler for tarball requests.

    This function also takes care of cleaning up old incomplete tarball constructions
    when given a named directory to do the construction in.

    The returned handler takes the hash and size_bytes of a Digest of a Directory
    message and constructs a tarball of the directory defined by the message.

    Args:
        context (Context): The context to use to send the gRPC requests.
        allow_all (bool): Whether or not to allow all CORS origins.
        allowed_origins (list): List of valid CORS origins.
        tarball_dir (str): Base directory to use for tarball construction.

    """

    class FetchSpec:
        """Simple class used to store information about a tarball request.

        A class is used here rather than a namedtuple since we need this state
        to be mutable.

        """

        def __init__(self, *, error: Optional[Exception], event: Optional[asyncio.Event], path: str, refcount: int):
            self.error = error
            self.event = event
            self.path = path
            self.refcount = refcount

    in_flight_fetches: Dict[str, FetchSpec] = {}
    fetch_lock = asyncio.Lock()

    # If we have a tarball directory to use, empty all existing tarball constructions from it
    # to provide some form of cleanup after a crash.
    if tarball_dir is not None:
        for path in os.listdir(tarball_dir):
            if path.startswith(TARBALL_DIRECTORY_PREFIX):
                shutil.rmtree(os.path.join(tarball_dir, path))

    async def _get_tarball_handler(request: web.Request) -> web.StreamResponse:
        digest_str = f'{request.match_info["hash"]}/{request.match_info["size_bytes"]}'
        LOGGER.info("Received request for a tarball from CAS for blob.", tags=dict(digest=digest_str))

        digest = Digest(hash=request.match_info["hash"], size_bytes=int(request.match_info["size_bytes"]))

        tmp_dir = tempfile.mkdtemp(prefix=TARBALL_DIRECTORY_PREFIX, dir=tarball_dir)

        try:
            duplicate_request = False
            event = None
            async with fetch_lock:
                if digest_str in in_flight_fetches:
                    LOGGER.info("Deduplicating request for tarball.", tags=dict(digest=digest_str))
                    spec = in_flight_fetches[digest_str]
                    spec.refcount += 1
                    event = spec.event
                    duplicate_request = True
                else:
                    event = asyncio.Event()
                    in_flight_fetches[digest_str] = FetchSpec(error=None, event=event, path="", refcount=1)

            if duplicate_request and event:
                # If we're a duplicate of an existing request, then wait for the
                # existing request to finish tarball creation before reading the
                # path from the cache.
                await event.wait()
                spec = in_flight_fetches[digest_str]
                if spec.error is not None:
                    raise spec.error
                tarball_path = in_flight_fetches[digest_str].path
            else:
                try:
                    directory = await _fetch_blob(context, cache, digest, message_class=Directory)
                    tarball_path = await _tarball_from_directory(context, cache, directory, tmp_dir)
                except web.HTTPError as e:
                    in_flight_fetches[digest_str].error = e
                    if event:
                        event.set()
                    raise e
                except Exception as e:
                    LOGGER.debug("Unexpected error constructing tarball.", tags=dict(digest=digest_str), exc_info=True)
                    in_flight_fetches[digest_str].error = e
                    if event:
                        event.set()
                    raise web.HTTPInternalServerError()

                # Update path in deduplication cache, and set event to notify
                # duplicate requests that the tarball is ready
                async with fetch_lock:
                    if event:
                        in_flight_fetches[digest_str].path = tarball_path
                        event.set()

            response = web.StreamResponse()

            # We need to explicitly set CORS headers here, because the middleware that
            # normally handles this only adds the header when the response is returned
            # from this function. However, when using a StreamResponse with chunked
            # encoding enabled, the client begins to receive the response when we call
            # `response.write()`. This leads to the request being disallowed due to the
            # missing reponse header for clients executing in browsers.
            cors_headers = get_cors_headers(request.headers.get("Origin"), allowed_origins, allow_all)
            response.headers.update(cors_headers)

            response.enable_chunked_encoding()
            await response.prepare(request)

            async with aiofiles.open(tarball_path, "rb") as tarball:
                await tarball.seek(0)
                chunk = await tarball.read(1024)
                while chunk:
                    await response.write(chunk)
                    chunk = await tarball.read(1024)
            return response

        except RpcError as e:
            LOGGER.warning(e.details())
            if e.code() == StatusCode.NOT_FOUND:
                raise web.HTTPNotFound()
            raise web.HTTPInternalServerError()

        finally:
            cleanup = False
            async with fetch_lock:
                # Decrement refcount now we're done with the tarball. If we're the
                # last request interested in the tarball then remove it along with
                # its construction directory.
                spec = in_flight_fetches[digest_str]
                spec.refcount -= 1
                if spec.refcount <= 0:
                    cleanup = True
                    in_flight_fetches.pop(digest_str)
            if cleanup:
                shutil.rmtree(tmp_dir)

    return _get_tarball_handler


def logstream_handler(context: Context) -> Callable[[web.Request], Awaitable[Any]]:
    async def _logstream_handler(request: web.Request) -> Any:
        LOGGER.info("Receieved request for a LogStream websocket.")
        stub = ByteStreamStub(context.logstream_channel)  # type: ignore # Requires stub regen
        ws = web.WebSocketResponse()
        await ws.prepare(request)

        async for msg in ws:
            if msg.type == WSMsgType.BINARY:
                read_request = ReadRequest()
                read_request.ParseFromString(msg.data)

                read_request.resource_name = f"{read_request.resource_name}"
                try:
                    async for response in stub.Read(read_request):
                        serialized_response = response.SerializeToString()
                        if serialized_response:
                            ws_response = {
                                "resource_name": read_request.resource_name,
                                "data": response.data.decode("utf-8"),
                                "complete": False,
                            }
                            await ws.send_json(ws_response)
                    ws_response = {"resource_name": read_request.resource_name, "data": "", "complete": True}
                    await ws.send_json(ws_response)
                except RpcError as e:
                    LOGGER.warning(e.details())
                    if e.code() == StatusCode.NOT_FOUND:
                        ws_response = {
                            "resource_name": read_request.resource_name,
                            "data": "NOT_FOUND",
                            "complete": True,
                        }
                        await ws.send_json(ws_response)
                    ws_response = {"resource_name": read_request.resource_name, "data": "INTERNAL", "complete": True}
                    await ws.send_json(ws_response)

    return _logstream_handler



# Copyright (C) 2024 Bloomberg LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  <http://www.apache.org/licenses/LICENSE-2.0>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Iterator

import pytest

from buildgrid._protos.build.bazel.remote.execution.v2.remote_execution_pb2 import ActionResult, Digest
from buildgrid.server.actioncache.caches.action_cache_abc import ActionCacheABC
from buildgrid.server.actioncache.caches.lru_cache import LruActionCache
from buildgrid.server.actioncache.caches.mirrored_cache import MirroredCache
from buildgrid.server.cas.storage import lru_memory_cache
from buildgrid.server.exceptions import NotFoundError


@pytest.fixture
def first_cache() -> ActionCacheABC:
    storage = lru_memory_cache.LRUMemoryCache(1024 * 1024)
    return LruActionCache(storage, max_cached_refs=1024)


@pytest.fixture
def second_cache() -> ActionCacheABC:
    storage = lru_memory_cache.LRUMemoryCache(1024 * 1024)
    return LruActionCache(storage, max_cached_refs=1024)


@pytest.fixture
def cache(first_cache: ActionCacheABC, second_cache: ActionCacheABC) -> Iterator[MirroredCache]:
    yield MirroredCache(first=first_cache, second=second_cache)


def test_mirrored_cache_both_missing(cache: ActionCacheABC):
    digest = Digest(hash="foo", size_bytes=123)
    with pytest.raises(NotFoundError):
        cache.get_action_result(digest)


def test_mirrored_first_missing(cache: ActionCacheABC, first_cache: ActionCacheABC, second_cache: ActionCacheABC):
    ar = ActionResult(exit_code=42)
    digest = Digest(hash="foo", size_bytes=123)
    second_cache.update_action_result(digest, ar)

    result = cache.get_action_result(digest)
    # copied into the first cache
    first_result = first_cache.get_action_result(digest)
    assert ar == result == first_result


def test_mirrored_second_missing(cache: ActionCacheABC, first_cache: ActionCacheABC, second_cache: ActionCacheABC):
    ar = ActionResult(exit_code=42)
    digest = Digest(hash="foo", size_bytes=123)
    first_cache.update_action_result(digest, ar)

    result = cache.get_action_result(digest)
    # copied into the second cache
    second_result = second_cache.get_action_result(digest)
    assert ar == result == second_result


def test_mirrored_both_existing_choose_first(
    cache: ActionCacheABC, first_cache: ActionCacheABC, second_cache: ActionCacheABC
):
    ar1 = ActionResult(exit_code=42)
    ar2 = ActionResult(exit_code=43)
    digest = Digest(hash="foo", size_bytes=123)
    first_cache.update_action_result(digest, ar1)
    second_cache.update_action_result(digest, ar2)

    result = cache.get_action_result(digest)
    assert ar1 == result
    assert ar2 != result


def test_mirrored_cache_update(cache: ActionCacheABC, first_cache: ActionCacheABC, second_cache: ActionCacheABC):
    ar = ActionResult(exit_code=42)
    digest = Digest(hash="foo", size_bytes=123)
    cache.update_action_result(digest, ar)

    assert ar == first_cache.get_action_result(digest)
    assert ar == second_cache.get_action_result(digest)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright (C) 2024 Bloomberg LP

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

 <http://www.apache.org/licenses/LICENSE-2.0>

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import abc
import buildgrid._protos.buildgrid.v2.introspection_pb2
import grpc

class IntrospectionStub:
    """BuildGrid's Introspection service is used to query the state of
    BuildGrid's internals, and obtain details on available functionality.
    """

    def __init__(self, channel: grpc.Channel) -> None: ...
    ListWorkers: grpc.UnaryUnaryMultiCallable[
        buildgrid._protos.buildgrid.v2.introspection_pb2.ListWorkersRequest,
        buildgrid._protos.buildgrid.v2.introspection_pb2.ListWorkersResponse,
    ]
    """Query the list of workers known to a specific BuildGrid instance

    Errors:

      * `INVALID_ARGUMENT`: When the provided instance name doesn't exist
    """
    GetOperationFilters: grpc.UnaryUnaryMultiCallable[
        buildgrid._protos.buildgrid.v2.introspection_pb2.GetOperationFiltersRequest,
        buildgrid._protos.buildgrid.v2.introspection_pb2.OperationFilters,
    ]
    """Get the list of valid filters for use in ListOperations requests
    for a specific BuildGrid instance

    Errors:

      * `INVALID_ARGUMENT`: When the provided instance name doesn't exist
    """

class IntrospectionServicer(metaclass=abc.ABCMeta):
    """BuildGrid's Introspection service is used to query the state of
    BuildGrid's internals, and obtain details on available functionality.
    """

    @abc.abstractmethod
    def ListWorkers(
        self,
        request: buildgrid._protos.buildgrid.v2.introspection_pb2.ListWorkersRequest,
        context: grpc.ServicerContext,
    ) -> buildgrid._protos.buildgrid.v2.introspection_pb2.ListWorkersResponse:
        """Query the list of workers known to a specific BuildGrid instance

        Errors:

          * `INVALID_ARGUMENT`: When the provided instance name doesn't exist
        """
    @abc.abstractmethod
    def GetOperationFilters(
        self,
        request: buildgrid._protos.buildgrid.v2.introspection_pb2.GetOperationFiltersRequest,
        context: grpc.ServicerContext,
    ) -> buildgrid._protos.buildgrid.v2.introspection_pb2.OperationFilters:
        """Get the list of valid filters for use in ListOperations requests
        for a specific BuildGrid instance

        Errors:

          * `INVALID_ARGUMENT`: When the provided instance name doesn't exist
        """

def add_IntrospectionServicer_to_server(servicer: IntrospectionServicer, server: grpc.Server) -> None: ...
